#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import struct
from io import BytesIO
from typing import TypeAlias

from PIL import Image

from xpra.log import Logger
from xpra.os_util import gi_import, POSIX, OSX
from xpra.util.io import load_binary_file
from xpra.util.objects import typedict
from xpra.gtk.window import add_close_accel
from xpra.gtk.signals import install_signal_handlers
from xpra.opengl.window import get_gl_client_window_module
from xpra.common import force_size_constraint
from xpra.client.gui.fake_client import FakeClient
from xpra.client.gui.window_border import WindowBorder

Gtk = gi_import("Gtk")
Pango = gi_import("Pango")

log = Logger()


ScreenUpdate: TypeAlias = tuple[int, int, int, int, str, bytes | tuple, int, typedict]


class TestGLRender:

    def __init__(self):
        from xpra.codecs.loader import load_codec
        load_codec("dec_pillow")

        self.ww = 800
        self.wh = 600

        self.index = 0
        self.updates = self.gen_updates()

        self.opengl_props, gl_client_window_module = get_gl_client_window_module("force")
        self.GLClientWindowClass = gl_client_window_module.GLClientWindow

        self.gl_window = self.init_gl_window()
        self.control_window = self.init_control_window()

    def gen_updates(self) -> list[ScreenUpdate]:
        # fake some draw packets
        w = self.ww
        h = self.wh

        def fullrgb(data: bytes) -> ScreenUpdate:
            return 0, 0, w, h, "rgb32", data, w * 4, typedict()

        def b(v: int) -> bytes:
            return struct.pack("@B", v)

        packets = [
            # clear it
            fullrgb(b"\0"*4*w*h),
            # white:
            fullrgb(b"\xff" * 4 * w * h),
            # gradient grey:
            fullrgb(b"".join(b(round(y * 255 / h)) * 4 * w for y in range(h))),
            # gradient grey:
            fullrgb(b"".join((b(round(y * 255 / h)) + b"\x80\x40\xff") * w for y in range(h)))
        ]

        def add_scroll(dy: int, fill: bytes):
            # scroll data: (x, y, w, h, xdelta, ydelta)
            scroll_h = abs(dy)
            scrolls = [
                (0, max(0, -dy), w, h-scroll_h, 0, dy)
            ]
            packets.append((0, 0, w, h, "scroll", scrolls, 0, typedict({"flush": 1})))
            stride = w * 4
            data = fill * stride * scroll_h
            if dy > 0:
                # scrolling down, fill the top:
                filly = 0
            else:
                filly = h + dy
            packets.append((0, filly, w, scroll_h, "rgb32", data, stride, typedict()))
        # scroll up:
        add_scroll(-10, b"\xff")
        # scroll down:
        add_scroll(-20, b"\x80")

        from xpra.platform.paths import get_icon_filename
        filename = get_icon_filename("xpra")
        ext = filename.split(".")[-1]
        icon = load_binary_file(filename)
        img = Image.open(BytesIO(icon))
        iw, ih = img.size
        assert icon

        def add_icon(x: int, y: int, options: typedict) -> None:
            packets.append((x, y, iw, ih, ext, icon, iw*4, options))
        count = h // ih + 1
        for y in range(count):
            add_icon(w // 2, y * ih, typedict({"flush": count-1-y}))
        return packets

    def init_control_window(self) -> None:
        window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
        window.set_transient_for(self.gl_window)
        window.set_size_request(1920, 720)

        vbox = Gtk.VBox(homogeneous=False, spacing=0)

        hbox = Gtk.HBox(homogeneous=True, spacing=10)
        vbox.pack_start(hbox, False, False, 10)

        pu = Gtk.Button(label="Previous Update")
        pu.connect("clicked", self.previous_update)
        hbox.add(pu)
        nu = Gtk.Button(label="Next Update")
        nu.connect("clicked", self.next_update)
        hbox.add(nu)

        self.label = Gtk.Label()
        fontdesc = Pango.FontDescription("monospace 9")
        self.label.modify_font(fontdesc)
        self.label.set_margin_start(10)
        self.label.set_xalign(0)
        self.label.set_line_wrap(True)
        vbox.pack_start(self.label, True, True, 0)

        window.add(vbox)

        window.connect("delete-event", self.close)
        add_close_accel(window, self.close)
        return window

    def init_gl_window(self) -> None:
        noclient = FakeClient()
        x = 2000
        y = 1000
        metadata = typedict({"has-alpha": True})
        metadata.update(force_size_constraint(self.ww, self.wh))
        border = WindowBorder()
        max_window_size = (4096, 4096)
        default_cursor_data = None
        pixel_depth = 32

        window = self.GLClientWindowClass(noclient, None, 0, 2 ** 32 - 1, x, y, self.ww, self.wh, self.ww, self.wh,
                                          metadata, False, typedict({}),
                                          border, max_window_size, default_cursor_data, pixel_depth)
        window_backing = window._backing
        window.realize()
        window_backing.paint_screen = True
        window.connect("delete-event", self.close)
        add_close_accel(window, self.close)
        return window

    def previous_update(self, *_args) -> None:
        if self.index > 0:
            self.index -= 1
            self.show_update()

    def next_update(self, *_args) -> None:
        if self.index < len(self.updates) - 1:
            self.index += 1
            self.show_update()

    def show_update(self) -> None:
        update = self.updates[self.index]

        def show_result(*args):
            log.info(f"show_result%s for {update[4]}", args)

        self.gl_window.draw_region(*update, callbacks=[show_result])
        encoding = update[4]
        x = update[0]
        y = update[1]
        self.label.set_text(f"{encoding} at {x}x{y}")

    def run(self) -> None:
        install_signal_handlers("test-gl-scroll", self.close)
        self.gl_window.show_all()
        self.control_window.show_all()
        self.show_update()
        Gtk.main()
        self.control_window.close()
        self.gl_window.close()

    def close(self, *_args) -> None:
        Gtk.main_quit()


def main() -> int:
    from xpra.platform import program_context
    from xpra.platform.gui import init as gui_init
    from xpra.log import enable_color, consume_verbose_argv
    with program_context("OpenGL Native Context Check"):
        if POSIX and not OSX:
            from xpra.x11.gtk.display_source import init_gdk_display_source
            init_gdk_display_source()
        gui_init()
        enable_color()
        consume_verbose_argv(sys.argv, "opengl")
        TestGLRender().run()
    return 0


if __name__ == "__main__":
    sys.exit(main())



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from time import monotonic
from collections.abc import Callable

from xpra.log import Logger
from xpra.os_util import POSIX
from xpra.util.env import envint

DISPLAY_FD_TIMEOUT = envint("XPRA_DISPLAY_FD_TIMEOUT", 20)


def write_displayfd(w_pipe, display, timeout: float = 10) -> int:
    import select
    import errno
    buf = ("%s\n" % display).encode("ascii")
    limit = monotonic() + timeout
    log = Logger("util")
    log("write_displayfd%s", (w_pipe, display, timeout))
    while buf and monotonic() < limit:
        try:
            timeout = max(0.0, limit - monotonic())
            if POSIX:
                w = select.select([], [w_pipe], [], timeout)[1]
                log("select.select(..) writeable=%s", w)
            else:
                w = [w_pipe]
            if w_pipe in w:
                count = os.write(w_pipe, buf)
                buf = buf[count:]
                log("wrote %i bytes, remains %s", count, buf)
        except OSError as e:
            if e.errno != errno.EINTR:
                raise
    if not buf:
        try:
            os.fsync(w_pipe)
        except OSError:
            log("os.fsync(%i)", w_pipe, exc_info=True)
        if w_pipe > 2:
            try:
                os.close(w_pipe)
            except OSError:
                log("os.close(%i)", w_pipe, exc_info=True)
    return len(buf) == 0


def read_displayfd(r_pipe, timeout=DISPLAY_FD_TIMEOUT, proc=None) -> bytes:
    import select
    import errno
    # Read the display number from the pipe we gave to Xvfb
    # waiting up to 10 seconds for it to show up
    limit = monotonic() + timeout
    buf = b""
    log = Logger("util")
    log("read_displayfd%s", (r_pipe, timeout, proc))
    while monotonic() < limit and len(buf) < 8 and (proc is None or proc.poll() is None):
        try:
            timeout = 1
            if POSIX:
                r = select.select([r_pipe], [], [], timeout)[0]
                log("readable=%s", r)
            else:
                r = [r_pipe]
            if r_pipe in r:
                v = os.read(r_pipe, 8)
                buf += v
                log("read=%s", v)
                if buf and (buf.endswith(b'\n') or len(buf) >= 8):
                    break
        except OSError as e:
            if e.errno != errno.EINTR:
                raise
    return buf


def parse_displayfd(buf: bytes, err: Callable) -> int:
    if not buf:
        err("did not provide a display number using displayfd")
        return -1
    if not buf.endswith(b"\n"):
        err("output not terminated by newline: %r" % buf)
        return -1
    buf = buf.rstrip(b"\n\r")
    try:
        n = int(buf)
    except ValueError:
        err("display number is not a valid number: %r" % buf)
        return -1
    if n < 0 or n >= 2 ** 16:
        err("provided an invalid display number: %s" % n)
    return n



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
from typing import Any

from xpra.platform import platform_import


class NOGLContext:
    def __init__(self, _alpha=False):
        raise NotImplementedError()


GLContext = NOGLContext


def check_support() -> dict[str, Any]:
    if not GLContext:
        raise RuntimeError("no GLContext available")
    return GLContext().check_support()  # pylint: disable=not-callable


platform_import(globals(), "gl_context", False, "GLContext", "check_support")


def main():
    from xpra.platform import program_context
    from xpra.platform.gui import init as gui_init
    from xpra.util.str_fn import print_nested_dict
    from xpra.log import enable_color, Logger, consume_verbose_argv
    with program_context("OpenGL Native Context Check"):
        gui_init()
        enable_color()
        consume_verbose_argv(sys.argv, "opengl")
        log = Logger("opengl")
        if not GLContext:
            log.error("Error: no GLContext available on %s", sys.platform)
            return 1
        try:
            props = check_support()
        except Exception:
            log.error("%s().check_support()", exc_info=True)
            return 1
        log.info("")
        log.info("OpenGL properties:")
        print_nested_dict(props)
        return 0


if __name__ == "__main__":
    sys.exit(main())



# This file is part of Xpra.
# Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any
from collections.abc import Sequence

from xpra.codecs.constants import preforder, STREAM_ENCODINGS
from xpra.codecs.loader import load_codec, codec_versions, has_codec, get_codec, unload_codecs
from xpra.codecs.video import getVideoHelper
from xpra.scripts.config import parse_bool_or_int
from xpra.common import FULL_INFO, VIDEO_MAX_SIZE, noop
from xpra.net.common import PacketType
from xpra.net import compression
from xpra.util.objects import typedict
from xpra.util.str_fn import csv
from xpra.util.env import envint, envbool
from xpra.client.base.stub_client_mixin import StubClientMixin
from xpra.log import Logger

log = Logger("client", "encoding")

B_FRAMES = envbool("XPRA_B_FRAMES", True)
PAINT_FLUSH = envbool("XPRA_PAINT_FLUSH", True)
MAX_SOFT_EXPIRED = envint("XPRA_MAX_SOFT_EXPIRED", 5)
SEND_TIMESTAMPS = envbool("XPRA_SEND_TIMESTAMPS", False)
SCROLL_ENCODING = envbool("XPRA_SCROLL_ENCODING", False)

# we assume that any server will support at least those:
DEFAULT_ENCODINGS = os.environ.get("XPRA_DEFAULT_ENCODINGS", "rgb32,rgb24,jpeg,png").split(",")


def get_core_encodings() -> Sequence[str]:
    """
        This method returns the actual encodings supported.
        ie: ["rgb24", "vp8", "webp", "png", "png/L", "png/P", "jpeg", "h264", "vpx"]
        It is often overridden in the actual client class implementations,
        where extra encodings can be added (generally just 'rgb32' for transparency),
        or removed if the toolkit implementation class is more limited.
    """
    # we always support rgb:
    core_encodings = ["rgb24", "rgb32"]
    for codec in ("dec_pillow", "dec_webp", "dec_jpeg", "dec_nvjpeg", "dec_avif"):
        if has_codec(codec):
            c = get_codec(codec)
            encs = c.get_encodings()
            log("%s.get_encodings()=%s", codec, encs)
            for e in encs:
                if e not in core_encodings:
                    core_encodings.append(e)
    if SCROLL_ENCODING:
        core_encodings.append("scroll")
    # we enable all the video decoders we know about,
    # but what will actually get used by the server will still depend
    # on the csc modes supported by the client
    video_decodings = getVideoHelper().get_decodings()
    log("video_decodings=%s", video_decodings)
    for encoding in video_decodings:
        if encoding not in core_encodings:
            core_encodings.append(encoding)
    # remove duplicates and use preferred encoding order:
    return preforder(core_encodings)


def get_batch_caps() -> dict[str, Any]:
    # batch options:
    caps = {}
    for bprop in ("always", "min_delay", "max_delay", "delay", "max_events", "max_pixels", "time_unit"):
        evalue = os.environ.get(f"XPRA_BATCH_{bprop.upper()}")
        if evalue:
            try:
                caps[f"batch.{bprop}"] = int(evalue)
            except ValueError:
                log.error("Error: invalid environment value for %s: %s", bprop, evalue)
    log("get_batch_caps()=%s", caps)
    return caps


class Encodings(StubClientMixin):
    """
    Mixin for adding encodings to a client
    """
    PREFIX = "encoding"

    def __init__(self):
        self.allowed_encodings = []
        self.encoding = ""
        self.quality = -1
        self.min_quality = 0
        self.speed = 0
        self.min_speed = -1
        self.video_scaling = None
        self.video_max_size = VIDEO_MAX_SIZE

        self.server_encodings: Sequence[str] = ()
        self.server_core_encodings: Sequence[str] = ()
        self.server_encodings_with_speed: Sequence[str] = ()
        self.server_encodings_with_quality: Sequence[str] = ()
        self.server_encodings_with_lossless_mode: Sequence[str] = ()

        # what we told the server about our encoding defaults:
        self.encoding_defaults = {}

    def init(self, opts) -> None:
        self.allowed_encodings = opts.encodings
        self.encoding = opts.encoding
        if opts.video_scaling.lower() in ("auto", "on"):
            self.video_scaling = None
        else:
            self.video_scaling = parse_bool_or_int("video-scaling", opts.video_scaling)
        self.quality = opts.quality
        self.min_quality = opts.min_quality
        self.speed = opts.speed
        self.min_speed = opts.min_speed
        load_codec("dec_pillow")
        ae = self.allowed_encodings
        if "png" in ae:
            # try to load the fast png decoder:
            load_codec("dec_spng")
        if "jpeg" in ae:
            # try to load the fast jpeg decoders:
            load_codec("dec_jpeg")
            load_codec("dec_nvjpeg")
            load_codec("nvdec")
        if "webp" in ae:
            # try to load the fast webp decoder:
            load_codec("dec_webp")
        if "avif" in ae:
            load_codec("dec_avif")
        vh = getVideoHelper()
        vh.set_modules(video_decoders=opts.video_decoders, csc_modules=opts.csc_modules)
        vh.init()

    def cleanup(self) -> None:
        with log.trap_error("Error during video helper cleanup"):
            getVideoHelper().cleanup()
        unload_codecs()

    def init_authenticated_packet_handlers(self) -> None:
        self.add_packets(f"{Encodings.PREFIX}-set")
        self.add_legacy_alias("encodings", f"{Encodings.PREFIX}-set")

    def _process_encoding_set(self, packet: PacketType) -> None:
        caps = typedict(packet[1])
        Encodings._parse_server_capabilities(self, caps)
        # fire setting change event for the system-tray:
        ssc = getattr(self, "server_setting_changed", noop)
        ssc("encoding", caps)

    def get_info(self) -> dict[str, Any]:
        return {
            "encodings": {
                "core": self.get_core_encodings(),
                "window-icon": self.get_window_icon_encodings(),
                "cursor": self.get_cursor_encodings(),
                "quality": self.quality,
                "min-quality": self.min_quality,
                "speed": self.speed,
                "min-speed": self.min_speed,
                "encoding": self.encoding or "auto",
                "video-scaling": self.video_scaling if self.video_scaling is not None else "auto",
            },
            "server-encodings": self.server_core_encodings,
        }

    def get_caps(self) -> dict[str, Any]:
        caps = {
            "encodings": {
                "": self.get_encodings(),
                "all": self.get_encodings(),
                "core": self.get_core_encodings(),
                "window-icon": self.get_window_icon_encodings(),
                "cursor": self.get_cursor_encodings(),
                "packet": True,
            },
            "batch": get_batch_caps(),
            "encoding": self.get_encodings_caps(),
        }
        return caps

    def parse_server_capabilities(self, c: typedict) -> bool:
        self._parse_server_capabilities(c)
        return True

    def _parse_server_capabilities(self, c) -> None:
        self.server_encodings = c.strtupleget("encodings", DEFAULT_ENCODINGS)
        self.server_core_encodings = c.strtupleget("encodings.core", self.server_encodings)
        # old servers only supported x264:
        self.server_encodings_with_speed = c.strtupleget("encodings.with_speed", ("h264",))
        self.server_encodings_with_quality = c.strtupleget("encodings.with_quality", ("jpeg", "webp", "h264"))
        self.server_encodings_with_lossless_mode = c.strtupleget("encodings.with_lossless_mode", ())
        e = c.strget("encoding")
        if e and not c.boolget("encodings.delayed"):
            if self.encoding and e != self.encoding:
                if self.encoding not in self.server_core_encodings:
                    log.warn("server does not support %s encoding and has switched to %s", self.encoding, e)
                else:
                    log.info("server is using %s encoding instead of %s", e, self.encoding)
            self.encoding = e

    def get_encodings_caps(self) -> dict[str, Any]:
        video_b_frames: list[str] = []
        if B_FRAMES:
            video_b_frames.append("h264")  # only tested with dec_avcodec2
        caps: dict[str, Any] = {
            "video_b_frames": video_b_frames,
            "video_max_size": self.video_max_size,
            "max-soft-expired": MAX_SOFT_EXPIRED,
            "send-timestamps": SEND_TIMESTAMPS,
        }
        if self.video_scaling is not None:
            caps["scaling.control"] = self.video_scaling
        if self.encoding:
            caps[""] = self.encoding
            # since v6.3:
            caps["setting"] = self.encoding
        if FULL_INFO > 1:
            for k, v in codec_versions.items():
                caps[f"{k}.version"] = v
        if self.quality > 0:
            caps["quality"] = self.quality
        if self.min_quality > 0:
            caps["min-quality"] = self.min_quality
        if self.speed >= 0:
            caps["speed"] = self.speed
        if self.min_speed >= 0:
            caps["min-speed"] = self.min_speed

        # generic rgb compression flags:
        if "lz4" in compression.get_enabled_compressors():
            caps["rgb_lz4"] = True
        # these are the defaults - when we instantiate a window,
        # we can send different values as part of the map event
        # these are the RGB modes we want (the ones we are expected to be able to paint with):
        rgb_formats = ["RGB", "RGBX", "RGBA"]
        caps["rgb_formats"] = rgb_formats
        # figure out which CSC modes (usually YUV) can give us those RGB modes:
        full_csc_modes = getVideoHelper().get_server_full_csc_modes_for_rgb(*rgb_formats)
        if has_codec("dec_webp"):
            full_csc_modes["webp"] = ["BGRX", "BGRA", "RGBX", "RGBA"] if self.opengl_enabled else ["BGRX", "BGRA"]
        if has_codec("dec_jpeg") or has_codec("dec_pillow"):
            full_csc_modes["jpeg"] = ["BGRX", "BGRA", "YUV420P"]
        if has_codec("dec_jpeg"):
            full_csc_modes["jpega"] = ["BGRA", "RGBA"]
        log("supported full csc_modes=%s", full_csc_modes)
        caps["full_csc_modes"] = full_csc_modes

        if "h264" in self.get_core_encodings():
            # some profile options: "baseline", "main", "high", "high10", ...
            # set the default to "high10" for YUV420P
            # as the python client always supports all the profiles
            # whereas on the server side, the default is baseline to accommodate less capable clients.
            # YUV422P requires high422, and
            # YUV444P requires high444,
            # so we don't bother specifying anything for those two.
            h264_caps = {}
            for csc_name, default_profile in (
                    ("YUV420P", "high"),
                    ("YUV422P", ""),
                    ("YUV444P", ""),
            ):
                profile = os.environ.get(f"XPRA_H264_{csc_name}_PROFILE", default_profile)
                if profile:
                    h264_caps[f"{csc_name}.profile"] = profile
            h264_caps["fast-decode"] = envbool("XPRA_X264_FAST_DECODE", False)
            log("x264 encoding options: %s", h264_caps)
            caps["h264"] = h264_caps
        caps["batch"] = get_batch_caps()
        log("encoding capabilities: %s", caps)
        return caps

    def get_encodings(self) -> Sequence[str]:
        """
            Unlike get_core_encodings(), this method returns "rgb" for both "rgb24" and "rgb32".
            That's because although we may support both, the encoding chosen is plain "rgb",
            and the actual encoding used ("rgb24" or "rgb32") depends on the window's bit depth.
            ("rgb32" if there is an alpha channel, and if the client supports it)
        """
        cenc = [{"rgb32": "rgb", "rgb24": "rgb"}.get(x, x) for x in self.get_core_encodings()]
        if "grayscale" not in cenc and "png/L" in cenc:
            cenc.append("grayscale")
        if any(x in cenc for x in STREAM_ENCODINGS):
            cenc.append("stream")
        return preforder(cenc)

    def get_cursor_encodings(self) -> list[str]:
        e = ["raw", "default"]
        if "png" in self.get_core_encodings():
            e.append("png")
        return e

    def get_window_icon_encodings(self) -> list[str]:
        e = ["BGRA", "default"]
        if "png" in self.get_core_encodings():
            e.append("png")
        return e

    def get_core_encodings(self) -> list[str]:
        core = get_core_encodings()
        r = [x for x in core if x in self.allowed_encodings]
        log(f"get_core_encodings()={r} (core={core}, allowed={self.allowed_encodings})")
        return r

    def set_encoding(self, encoding: str) -> None:
        log("set_encoding(%s)", encoding)
        if encoding != "auto":
            encodings = self.get_encodings()
            if encoding not in encodings:
                raise ValueError(f"encoding {encoding} is not supported - only {csv(encodings)!r}")
            if encoding not in self.server_encodings:
                log.error(f"Error: encoding {encoding} is not supported by the server")
                log.error(" the only encodings allowed are:")
                log.error(" " + csv(self.server_encodings))
                return
        self.encoding = encoding
        self.send("encoding", self.encoding)

    def send_quality(self) -> None:
        q = self.quality
        log("send_quality() quality=%s", q)
        if q != -1 and (q < 0 or q > 100):
            raise ValueError(f"invalid quality: {q}")
        self.send("quality", q)

    def send_min_quality(self) -> None:
        q = self.min_quality
        log("send_min_quality() min-quality=%s", q)
        if q != -1 and (q < 0 or q > 100):
            raise ValueError(f"invalid min-quality: {q}")
        self.send("min-quality", q)

    def send_speed(self) -> None:
        s = self.speed
        log("send_speed() min-speed=%s", s)
        if s != -1 and (s < 0 or s > 100):
            raise ValueError(f"invalid speed: {s}")
        self.send("speed", s)

    def send_min_speed(self) -> None:
        s = self.min_speed
        log("send_min_speed() min-speed=%s", s)
        if s != -1 and (s < 0 or s > 100):
            raise ValueError(f"invalid min-speed: {s}")
        self.send("min-speed", s)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import json
import glob
import os.path
import binascii
import base64
from struct import pack, unpack
from hashlib import sha256
from collections.abc import Sequence

from xpra.util.objects import typedict
from xpra.util.str_fn import csv, strtobytes, hexstr
from xpra.os_util import getuid, POSIX
from xpra.util.env import osexpand
from xpra.util.io import load_binary_file
from xpra.net.digest import get_salt
from xpra.auth.sys_auth_base import SysAuthenticator, log
from xpra.platform.paths import get_user_conf_dirs

PUB_KEY_DER_PREFIX = binascii.a2b_hex("3059301306072a8648ce3d020106082a8648ce3d030107034200")
APP_ID = os.environ.get("XPRA_U2F_APP_ID", "Xpra")


class Authenticator(SysAuthenticator):

    def __init__(self, **kwargs):
        self.app_id = kwargs.pop("app_id", APP_ID)
        key_hexstring = kwargs.pop("public_key", "")
        super().__init__(**kwargs)
        self.public_keys = {}
        key_strs = {}
        if key_hexstring:
            log("u2f_auth: public key from configuration=" + key_hexstring)
            key_strs["command-option"] = key_hexstring
        # try to load public keys from the user conf dir(s):
        if getuid() == 0 and POSIX:
            # root: use the uid of the username specified:
            uid = self.get_uid()
        else:
            uid = getuid()
        conf_dirs = get_user_conf_dirs(uid)
        log("u2f: will try to load public keys from " + csv(conf_dirs))
        # load public keys:
        for d in conf_dirs:
            ed = osexpand(d)
            if os.path.exists(ed) and os.path.isdir(ed):
                pub_keyfiles = glob.glob(os.path.join(ed, "u2f*-pub.hex"))
                log(f"u2f: keyfiles({ed})={pub_keyfiles}")
                for f in sorted(pub_keyfiles):
                    key_hexstring = load_binary_file(f)
                    if key_hexstring:
                        key_hexstring = key_hexstring.rstrip(b" \n\r")
                        key_strs[f] = key_hexstring
                        log(f"u2f_auth: loaded public key from file {f!r}: {key_hexstring}")
        # parse public key data:
        # pylint: disable=import-outside-toplevel
        from cryptography.hazmat.primitives.serialization import load_der_public_key
        from cryptography.hazmat.backends import default_backend
        for origin, key_hexstring in key_strs.items():
            try:
                key = binascii.unhexlify(key_hexstring)
            except Exception as e:
                log(f"unhexlify({key_hexstring})", exc_info=True)
                log.warn(f"Warning: failed to parse key {origin!r}")
                log.warn(f" {e}")
                continue
            log(f"u2f: trying to load DER public key {key!r}")
            if not key.startswith(PUB_KEY_DER_PREFIX):
                key = PUB_KEY_DER_PREFIX + key
            try:
                k = load_der_public_key(key, default_backend())
            except Exception as e:
                log("load_der_public_key(%r)", key, exc_info=True)
                log.warn(f"Warning: failed to parse key {origin!r}")
                log.warn(f" {e}")
                continue
            self.public_keys[origin] = k
        if not self.public_keys:
            raise RuntimeError("u2f authenticator requires at least one public key")
        self.authenticate_check = self.u2f_check

    def get_challenge(self, digests: Sequence[str]) -> tuple[bytes, str]:
        if "u2f" not in digests:
            log.error("Error: client does not support u2f authentication")
            return b"", ""
        self.salt = get_salt()
        self.digest = "u2f:xor"
        self.challenge_sent = True
        return self.salt, self.digest

    def __repr__(self):
        return "u2f"

    def u2f_check(self, caps: typedict) -> bool:
        challenge_response = caps.strget("challenge_response")
        client_salt = caps.strget("challenge_client_salt")
        log(f"authenticate_check: response={challenge_response}, client-salt={client_salt}")
        user_presence, counter = unpack(b">BI", strtobytes(challenge_response)[:5])
        sig = strtobytes(challenge_response[5:])
        log(f"u2f user_presence={user_presence}, counter={counter}, signature={hexstr(sig)}")
        app_param = sha256(self.app_id.encode('utf8')).digest()
        server_challenge_b64 = base64.urlsafe_b64encode(self.salt).decode()
        server_challenge_b64 = server_challenge_b64.rstrip('=')
        log("challenge_b64(%s)=%s", repr(self.salt), server_challenge_b64)
        client_data = {
            "challenge": server_challenge_b64,
            "origin": client_salt,
            "typ": "navigator.id.getAssertion",
        }
        client_param = sha256(json.dumps(client_data, sort_keys=True).encode('utf8')).digest()
        param = app_param + pack(b'>B', user_presence) + pack(b'>I', counter) + client_param
        # check all the public keys:
        # pylint: disable=import-outside-toplevel
        from cryptography.hazmat.primitives import hashes
        from cryptography.hazmat.primitives.asymmetric import ec
        errors = {}
        for origin, public_key in self.public_keys.items():
            try:
                public_key.verify(sig, param, ec.ECDSA(hashes.SHA256()))
                log(f"ECDSA SHA256 verification passed for {origin!r}")
                return True
            except Exception as e:
                log(f"authenticate failed for {origin!r} / {public_key}", exc_info=True)
                errors[origin] = str(e) or type(e)
        log.error("Error: authentication failed,")
        log.error(f" checked against {len(self.public_keys)} keys")
        for origin, error in errors.items():
            log.error(f" {origin!r}: {error}")
        return False



#!/usr/bin/env python3

import gi

gi.require_version("Gtk", "3.0")  # @UndefinedVariable
from gi.repository import Gtk    # pylint: disable=wrong-import-position @UnresolvedImport
from xpra.gtk.util import get_root_size

width = 400
height = 200
def main():
	window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
	window.set_size_request(width, height)
	window.connect("delete_event", Gtk.main_quit)
	vbox = Gtk.VBox(homogeneous=False, spacing=0)
	hbox = Gtk.HBox(homogeneous=False, spacing=0)
	vbox.pack_start(hbox, expand=False, fill=False, padding=10)
	btn = Gtk.Button("move me")
	hbox.pack_start(btn, expand=False, fill=False, padding=10)
	def move(*_args):
		x, y = window.get_position()
		maxx, maxy = get_root_size()
		new_x = (x+100) % maxx
		new_y = (y+100) % maxy
		print("moving to %s x %s" % (new_x, new_y))
		window.move(new_x, new_y)
	btn.connect('clicked', move)
	window.add(vbox)
	window.show_all()
	Gtk.main()
	return 0


if __name__ == "__main__":
	main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.platform.posix.fd_portal_shadow import PortalShadow
from xpra.log import Logger

log = Logger("shadow")


class ScreenCast(PortalShadow):

    def get_server_mode(self) -> str:
        return "pipewire screencast"

    def on_session_created(self) -> None:
        # skip select_devices() and go straight to sources then start:
        self.select_sources()

    def set_keymap(self, server_source, force=False) -> None:
        """
        no input devices
        """

    def do_process_button_action(self, *args) -> None:
        """
        no input devices
        """



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

# pylint: disable=import-outside-toplevel


class TestNVUtil(unittest.TestCase):

    def test_nvutil(self) -> None:
        from xpra.codecs.nvidia.util import get_nvml_driver_version, get_proc_driver_version
        v1 = get_nvml_driver_version()
        v2 = get_proc_driver_version()
        if v1 and v2:
            assert v1 == v2, f"versions differ: {v1} (nvml) and {v2} (proc)"


def main():
    try:
        from xpra.codecs.nvidia import util
        assert util
    except ImportError:
        print("nvidia.util codec not installed - test skipped")
    else:
        unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os

from xpra.util.env import osexpand, envbool


LOG_EOF = envbool("XPRA_SSH_LOG_EOF", True)


def get_default_keyfiles() -> list[str]:
    dkf = os.environ.get("XPRA_SSH_DEFAULT_KEYFILES", None)
    if dkf is not None:
        return [x for x in dkf.split(os.pathsep) if x]
    return [osexpand(os.path.join("~/", ".ssh", keyfile)) for keyfile in ("id_ed25519", "id_ecdsa", "id_rsa", "id_dsa")]



# This file is part of Xpra.
# Copyright (C) 2025 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Dict

from libc.stddef cimport wchar_t
from libc.stdint cimport uint64_t, uintptr_t

from xpra.codecs.amf.amf cimport (
    AMF_RESULT, AMFFactory, AMFGuid, AMFSurface, AMFSurfaceVtbl,
    AMFPlane, AMFPlaneVtbl, AMFData, AMFCaps, AMFIOCaps, AMFVariantStruct,
    amf_uint32, amf_uint16, amf_uint8,
)


cdef AMFFactory *get_factory()
cdef void cleanup(self)
cdef object get_version()

cdef void check(AMF_RESULT res, message)
cdef object error_str(AMF_RESULT result)

cdef void set_guid(AMFGuid *guid,
                   amf_uint32 _data1, amf_uint16 _data2, amf_uint16 _data3,
                   amf_uint8 _data41, amf_uint8 _data42, amf_uint8 _data43, amf_uint8 _data44,
                   amf_uint8 _data45, amf_uint8 _data46, amf_uint8 _data47, amf_uint8 _data48)

cdef uint64_t get_c_version()

cdef void fill_nv12_surface(AMFSurface *surface, amf_uint8 Y, amf_uint8 U, amf_uint8 V)

cdef object get_caps(AMFCaps *caps, props: Dict)

cdef object get_io_caps(AMFIOCaps *iocaps)

cdef object get_plane_info(AMFPlane *plane)

cdef object get_data_info(AMFData *data)

cdef object get_surface_info(AMFSurface *surface)



# This file is part of Xpra.
# Copyright (C) 2025 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any
from collections.abc import Callable

from xpra.util.str_fn import Ellipsizer
from xpra.log import Logger

log = Logger("ibus")

BUSNAME = "org.freedesktop.IBus"


def noemptyvalues(d: dict) -> dict:
    return dict((k, v) for k, v in d.items() if v)


def query_engine(e) -> dict[str, Any]:
    return noemptyvalues({
        "description": e.get_description(),
        "language": e.get_language(),
        "layout": e.get_layout(),
        "option": e.get_layout_option(),
        "variant": e.get_layout_variant(),
        "long-name": e.get_longname(),
        "name": e.get_name(),
        "symbol": e.get_symbol(),
        "version": e.get_version(),
        "rank": e.get_rank(),
    })


def query_ibus() -> dict[str, Any]:
    try:
        from xpra.os_util import gi_import
        IBus = gi_import("IBus")
    except ImportError:
        return {}
    bus = IBus.Bus()
    info = {
        "address": IBus.get_address(),
        "machine-id": IBus.get_local_machine_id(),
        "socket": IBus.get_socket_path(),
        "connected": bus.is_connected(),
    }
    if bus.is_connected():
        ge = bus.get_global_engine()
        if ge:
            info["engine"] = query_engine(ge)
        engines = bus.list_engines()
        if engines:
            info["engines"] = tuple(query_engine(engine) for engine in engines)
    log("query_ibus()=%s", Ellipsizer(info))
    return info


def with_ibus_ready(callback: Callable[[], None]):
    try:
        from xpra.os_util import gi_import
        from xpra.dbus.helper import DBusHelper
        IBus = gi_import("IBus")
        dbus_helper = DBusHelper()
    except ImportError as e:
        log(f"failed to import ibus functions: {e}")
        return

    # warning: this callback will not fire unless a GLib main loop is running!
    def got_name(name) -> None:
        log("got_name(%r) for %s", name, BUSNAME)
        if not name:
            return
        bus = IBus.Bus()

        def check_connected(delay: int) -> bool:
            connected = bus.is_connected()
            log("IBus.Bus.connected()=%s", connected)
            if connected:
                callback()
                return False
            if delay >= 10000:
                log.warn("Warning: timeout waiting for IBus")
                return False

            delay += 1000
            GLib = gi_import("GLib")
            GLib.timeout_add(delay, check_connected, delay)
            return False
        check_connected(0)

    log(f"waiting for {BUSNAME!r} to call {callback}")
    session_bus = dbus_helper.get_session_bus()
    session_bus.watch_name_owner(BUSNAME, got_name)


def set_engine(name: str) -> bool:
    log(f"ibus.set_engine({name!r})")
    try:
        from xpra.os_util import gi_import
        IBus = gi_import("IBus")
    except ImportError as e:
        log(f"failed to import ibus: {e}")
        return False
    bus = IBus.Bus()
    if not bus.is_connected():
        log(f"bus {bus} is not connected")
        return False
    r = bus.set_global_engine(name)
    log("%s.set_global_engine(%s)=%s", bus, name, r)
    return r


def get_engine_layout_spec() -> tuple[str, str, str]:
    try:
        from xpra.os_util import gi_import
        IBus = gi_import("IBus")
    except ImportError as e:
        log(f"failed to import ibus: {e}")
        return "", "", ""
    bus = IBus.Bus()
    if not bus.is_connected():
        log(f"bus {bus} is not connected")
        return "", "", ""
    engine = bus.get_global_engine()
    if not engine:
        return "", "", ""
    return engine.get_layout(), engine.get_layout_variant(), engine.get_layout_option()


def main(_argv) -> int:  # pragma: no cover
    from xpra.util.str_fn import print_nested_dict
    info = query_ibus()
    engines = info.pop("engines", ())
    print_nested_dict(info)
    print_nested_dict({"engines": dict(enumerate(engines))})
    return 0


if __name__ == "__main__":
    import sys
    main(sys.argv)



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys

from xpra.util.env import envbool
from xpra.notifications.notifier_base import NotifierBase, log, NID
from xpra.platform.win32.balloon import notify

GTK_NOTIFIER = envbool("XPRA_WIN32_GTK_NOTIFIER", False)
if GTK_NOTIFIER:
    try:
        from xpra.gtk.notifier import GTKNotifier
    except ImportError:
        GTK_NOTIFIER = False


def do_notify(*args):
    # if GLib is loaded, use it to ensure we use the UI thread:
    GLib = sys.modules.get("gi.repository.GLib", None)
    if GLib:
        GLib.idle_add(notify, *args)
    else:
        notify(*args)


class Win32_Notifier(NotifierBase):

    def __init__(self, *args):
        super().__init__(*args)
        self.handles_actions = GTK_NOTIFIER
        self.gtk_notifier = None
        self.gtk_notifications = set()
        self.notification_handles = {}

    def get_gtk_notifier(self):
        if self.gtk_notifier is None and GTK_NOTIFIER:
            try:
                self.gtk_notifier = GTKNotifier(self.closed_cb, self.action_cb)
            except Exception:
                log("failed to load GTK Notifier fallback", exc_info=True)
        return self.gtk_notifier

    def show_notify(self, dbus_id, tray, nid: NID,
                    app_name: str, replaces_nid: NID,
                    app_icon, summary: str, body: str, actions, hints, expire_timeout: int, icon):
        if not tray:
            log.warn("Warning: cannot show notifications without a system tray")
            return
        getHWND = getattr(tray, "getHWND", None)
        if GTK_NOTIFIER and (actions or not getHWND):
            log("show_notify(..) using gtk fallback, GTK_NOTIFIER=%s, tray=%s, getHWND=%s, actions=%s",
                GTK_NOTIFIER, tray, getHWND, actions)
            gtk_notifier = self.get_gtk_notifier()
            if gtk_notifier:
                gtk_notifier.show_notify(dbus_id, tray, nid, app_name, replaces_nid, app_icon, summary, body, actions,
                                         hints, expire_timeout, icon)
                self.gtk_notifications.add(nid)
                return
        if not getHWND:
            log.warn(f"Warning: missing 'getHWND' on {tray} ({type(tray)}")
            return
        if tray is None:
            log.warn("Warning: no system tray - cannot show notification!")
            return
        hwnd = getHWND()
        app_id = tray.app_id
        log("show_notify%s hwnd=%#x, app_id=%i",
            (dbus_id, tray, nid, app_name, replaces_nid, app_icon, summary, body, actions, hints, expire_timeout, icon),
            hwnd, app_id)
        # FIXME: remove handles when notification is closed
        self.notification_handles[int(nid)] = (hwnd, app_id)
        do_notify(hwnd, app_id, summary, body, expire_timeout, icon)

    def close_notify(self, nid: NID):
        try:
            self.gtk_notifications.remove(int(nid))
            if self.gtk_notifier:
                self.gtk_notifier.close_notify(nid)
        except KeyError:
            try:
                hwnd, app_id = self.notification_handles.pop(int(nid))
            except KeyError:
                return
            log("close_notify(%i) hwnd=%i, app_id=%i", nid, hwnd, app_id)
            do_notify(hwnd, app_id, "", "", 0, None)



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Final
from math import ceil, floor

from xpra.os_util import gi_import
from xpra.util.objects import typedict
from xpra.util.env import envint, envbool
from xpra.common import MAX_WINDOW_SIZE
from xpra.gtk.gobject import one_arg_signal
from xpra.gtk.error import XError, xsync, xswallow, xlog
from xpra.x11.gtk.prop import prop_set
from xpra.x11.prop_conv import MotifWMHints
from xpra.x11.bindings.window import X11WindowBindings
from xpra.x11.bindings.send_wm import send_wm_take_focus
from xpra.x11.common import Unmanageable
from xpra.x11.models.size_hints_util import sanitize_size_hints
from xpra.x11.models.base import BaseWindowModel, constants
from xpra.x11.models.core import sanestr
from xpra.x11.gtk.bindings import add_event_receiver, remove_event_receiver
from xpra.log import Logger

log = Logger("x11", "window")
workspacelog = Logger("x11", "window", "workspace")
shapelog = Logger("x11", "window", "shape")
grablog = Logger("x11", "window", "grab")
metalog = Logger("x11", "window", "metadata")
iconlog = Logger("x11", "window", "icon")
focuslog = Logger("x11", "window", "focus")
geomlog = Logger("x11", "window", "geometry")

GObject = gi_import("GObject")
GLib = gi_import("GLib")

X11Window = X11WindowBindings()

CWX: Final[int] = constants["CWX"]
CWY: Final[int] = constants["CWY"]
CWWidth: Final[int] = constants["CWWidth"]
CWHeight: Final[int] = constants["CWHeight"]
CWBorderWidth: Final[int] = constants["CWBorderWidth"]
CWSibling: Final[int] = constants["CWSibling"]
CWStackMode: Final[int] = constants["CWStackMode"]
CONFIGURE_GEOMETRY_MASK: Final[int] = CWX | CWY | CWWidth | CWHeight
CW_MASK_TO_NAME: dict[int, str] = {
    CWX: "X",
    CWY: "Y",
    CWWidth: "Width",
    CWHeight: "Height",
    CWSibling: "Sibling",
    CWStackMode: "StackMode",
    CWBorderWidth: "BorderWidth",
}


def configure_bits(value_mask: int) -> str:
    return "|".join(v for k, v in CW_MASK_TO_NAME.items() if k & value_mask)


FORCE_XSETINPUTFOCUS = envbool("XPRA_FORCE_XSETINPUTFOCUS", False)
VALIDATE_CONFIGURE_REQUEST = envbool("XPRA_VALIDATE_CONFIGURE_REQUEST", False)
CLAMP_OVERLAP = envint("XPRA_WINDOW_CLAMP_OVERLAP", 20)
assert CLAMP_OVERLAP >= 0
SIZEHINTS_POSITION = envbool("XPRA_SIZEHINTS_POSITION", False)
SIZEHINTS_SIZE = envbool("XPRA_SIZEHINTS_SIZE", False)


def calc_constrained_size(width: int, height: int, hints) -> tuple[int, int]:
    if not hints:
        return width, height

    def getintpair(key: str, dv1: int, dv2: int) -> tuple[int, int]:
        v = hints.get(key)
        if v:
            try:
                return int(v[0]), int(v[1])
            except (ValueError, IndexError):
                pass
        return dv1, dv2

    min_width: int
    min_height: int
    max_width: int
    max_height: int
    base_width: int
    base_height: int
    e_width: int
    e_height: int
    increment_x: int
    increment_y: int

    min_width, min_height = getintpair("minimum-size", 0, 0)
    if min_width > 0:
        width = max(min_width, width)
    if min_height > 0:
        height = max(min_height, height)

    max_width, max_height = getintpair("maximum-size", 2 ** 16 - 1, 2 ** 16 - 1)
    if max_width > 0:
        width = min(max_width, width)
    if max_height > 0:
        height = min(max_height, height)

    base_width, base_height = getintpair("base-size", 0, 0)
    increment_x, increment_y = getintpair("increment", 0, 0)
    if increment_x:
        e_width = max(width - base_width, 0)
        width -= e_width % increment_x
    if increment_y:
        e_height = max(height - base_height, 0)
        height -= e_height % increment_y

    min_aspect: float
    max_aspect: float
    if "min_aspect" in hints:
        min_aspect = hints.get("min_aspect")
        assert min_aspect > 0
        if width / height < min_aspect:
            height = ceil(width * min_aspect)
            if increment_y > 1 or base_height > 0:
                e_height = max(height - base_height, 0)
                height += increment_y - (e_height % increment_y)
    if "max_aspect" in hints:
        max_aspect = hints.get("max_aspect")
        assert max_aspect > 0
        if width / height > max_aspect:
            height = floor(width * max_aspect)
            if increment_y > 1 or base_height > 0:
                e_height = max(height - base_height, 0)
                height = max(1, height - e_height % increment_y)
    return width, height


class WindowModel(BaseWindowModel):
    """This represents a managed client window.  It allows one to produce
    widgets that view that client window in various ways."""

    _NET_WM_ALLOWED_ACTIONS = [f"_NET_WM_ACTION_{x}" for x in (
        "CLOSE", "MOVE", "RESIZE", "FULLSCREEN",
        "MINIMIZE", "SHADE", "STICK",
        "MAXIMIZE_HORZ", "MAXIMIZE_VERT",
        "CHANGE_DESKTOP", "ABOVE", "BELOW")]

    __gproperties__ = dict(BaseWindowModel.__common_properties__)
    __gproperties__ |= {
        # Client(s) state, is this window shown somewhere?
        # (then it can't be iconified)
        "shown": (
            GObject.TYPE_BOOLEAN,
            "Is this window shown somewhere", "",
            False,
            GObject.ParamFlags.READWRITE,
        ),
        # To prevent resizing loops,
        # the client must provide this counter when resizing the window
        # (and if the counter has been incremented since, we ignore the request)
        "resize-counter": (
            GObject.TYPE_INT64,
            "", "",
            0, GLib.MAXINT64, 1,
            GObject.ParamFlags.READWRITE,
        ),
        "client-geometry": (
            GObject.TYPE_PYOBJECT,
            "The current window geometry used by client(s)", "",
            GObject.ParamFlags.READWRITE,
        ),
        # Interesting properties of the client window, that will be
        # automatically kept up to date:
        "requested-position": (
            GObject.TYPE_PYOBJECT,
            "Client-requested position on screen", "",
            GObject.ParamFlags.READABLE,
        ),
        "requested-size": (
            GObject.TYPE_PYOBJECT,
            "Client-requested size on screen", "",
            GObject.ParamFlags.READABLE
        ),
        "set-initial-position": (
            GObject.TYPE_BOOLEAN,
            "Should the requested position be honoured?", "",
            False,
            GObject.ParamFlags.READWRITE,
        ),
        # Toggling this property does not actually make the window `iconified`,
        # i.e. make it appear or disappear from the screen -- it merely
        # updates the various window manager properties that inform the world
        # whether or not the window is `iconified`.
        "iconic": (
            GObject.TYPE_BOOLEAN,
            "ICCCM 'iconic' state -- any sort of 'not on desktop'.", "",
            False,
            GObject.ParamFlags.READWRITE,
        ),
        # from WM_NORMAL_HINTS
        "size-hints": (
            GObject.TYPE_PYOBJECT,
            "Client hints on constraining its size", "",
            GObject.ParamFlags.READABLE,
        ),
        # from _NET_WM_ICON_NAME or WM_ICON_NAME
        "icon-title": (
            GObject.TYPE_PYOBJECT,
            "Icon title (unicode or None)", "",
            GObject.ParamFlags.READABLE,
        ),
        # from _NET_WM_ICON
        "icons": (
            GObject.TYPE_PYOBJECT,
            "Icons in raw RGBA format, by size", "",
            GObject.ParamFlags.READABLE,
        ),
        # from _MOTIF_WM_HINTS.decorations
        "decorations": (
            GObject.TYPE_INT,
            "Should the window decorations be shown", "",
            -1, 65535, -1,
            GObject.ParamFlags.READABLE,
        ),
        "children": (
            GObject.TYPE_PYOBJECT,
            "Sub-windows", None,
            GObject.ParamFlags.READABLE,
        ),
    }
    __gsignals__ = dict(BaseWindowModel.__common_signals__)
    __gsignals__ |= {
        "x11-child-map-request-event": one_arg_signal,
        "x11-child-configure-request-event": one_arg_signal,
        "x11-destroy-event": one_arg_signal,
    }

    _property_names = BaseWindowModel._property_names + [
        "size-hints", "icon-title", "icons", "decorations",
        "modal", "set-initial-position", "requested-position",
        "iconic",
    ]
    _dynamic_property_names = BaseWindowModel._dynamic_property_names + [
        "size-hints", "icon-title", "icons", "decorations", "modal", "iconic",
    ]
    _initial_x11_properties = BaseWindowModel._initial_x11_properties + [
        "WM_HINTS", "WM_NORMAL_HINTS", "_MOTIF_WM_HINTS",
        "WM_ICON_NAME", "_NET_WM_ICON_NAME", "_NET_WM_ICON",
        "_NET_WM_STRUT", "_NET_WM_STRUT_PARTIAL",
    ]
    _internal_property_names = BaseWindowModel._internal_property_names + ["children"]
    _MODELTYPE = "Window"

    def __init__(self, parking_window_xid: int, xid: int, desktop_geometry, size_constraints=None):
        """Register a new client window with the WM.

        Raises an Unmanageable exception if this window should not be
        managed, for whatever reason.  ATM, this mostly means that the window
        died somehow before we could do anything with it."""

        super().__init__(xid)
        self.parking_window_xid = parking_window_xid
        self.corral_xid: int = 0
        self.desktop_geometry = desktop_geometry
        self.size_constraints = size_constraints or (0, 0, MAX_WINDOW_SIZE, MAX_WINDOW_SIZE)
        self.saved_events = -1
        # these extra state attributes allow us to `unmanage()` the window cleanly:
        self.in_save_set: bool = False
        self.client_reparented: bool = False
        self.kill_count: int = 0

        self.call_setup()

    #########################################
    # Setup and teardown
    #########################################

    def setup(self) -> None:
        super().setup()
        self.saved_events = X11Window.getEventMask(self.xid)
        ogeom = X11Window.getGeometry(self.xid)
        if not ogeom:
            raise XError("window disappeared")
        ox, oy, ow, oh = ogeom[:4]
        # clamp this window to the desktop size:
        x, y = self._clamp_to_desktop(ox, oy, ow, oh)
        geomlog("setup() clamp_to_desktop(%s)=%s", ogeom, (x, y))
        self.corral_xid = X11Window.CreateCorralWindow(self.parking_window_xid, self.xid, x, y)
        log("setup() corral_xid=%#x", self.corral_xid)
        prop_set(self.corral_xid, "_NET_WM_NAME", "utf8", "Xpra-CorralWindow-%#x" % self.xid)
        X11Window.substructureRedirect(self.corral_xid)
        add_event_receiver(self.corral_xid, self)

        # The child might already be mapped, in case we inherited it from
        # a previous window manager.  If so, we unmap it now, and save the
        # serial number of the request -- this way, when we get an
        # UnmapNotify later, we'll know that it's just from us unmapping
        # the window, not from the client withdrawing the window.
        if X11Window.is_mapped(self.xid):
            log("hiding inherited window")
            self.last_unmap_serial = X11Window.Unmap(self.xid)

        log("setup() adding to save set")
        X11Window.XAddToSaveSet(self.xid)
        self.in_save_set = True

        log("setup() reparenting")
        X11Window.Reparent(self.xid, self.corral_xid, 0, 0)
        self.client_reparented = True

        geom = X11Window.geometry_with_border(self.xid)
        if geom is None:
            raise Unmanageable(f"window {self.xid:x} disappeared already")
        geomlog("setup() geometry=%s, ogeom=%s", geom, ogeom)
        nx, ny, w, h = geom[:4]
        # after reparenting, the coordinates of the client window should be 0,0
        # use the coordinates of the corral window:
        if nx == ny == 0:
            nx, ny = x, y
        hints = self.get_property("size-hints")
        geomlog("setup() hints=%s size=%ix%i", hints, w, h)
        nw, nh = self.calc_constrained_size(w, h, hints)
        self._updateprop("geometry", (nx, ny, nw, nh))
        geomlog("setup() resizing windows to %sx%s, moving to %i,%i", nw, nh, nx, ny)
        # don't trigger a move or resize unless we have to:
        moved = ox != nx or oy != ny
        resized = ow != nw or oh != nh
        if moved and resized:
            X11Window.MoveResizeWindow(self.corral_xid, nx, ny, nw, nh)
            X11Window.MoveResizeWindow(self.xid, 0, 0, nw, nh)
        elif moved:
            X11Window.MoveWindow(self.corral_xid, nx, ny)
        if not X11Window.is_mapped(self.xid):
            X11Window.MapWindow(self.xid)
        # this is here to trigger X11 errors if any are pending
        # or if the window is deleted already:
        assert X11Window.getGeometry(self.xid)
        self._internal_set_property("shown", False)
        self._internal_set_property("resize-counter", 0)
        self._internal_set_property("client-geometry", None)

    def _clamp_to_desktop(self, x: int, y: int, w: int, h: int) -> tuple[int, int]:
        if self.desktop_geometry:
            dw, dh = self.desktop_geometry
            if x + w < 0:
                x = min(0, CLAMP_OVERLAP - w)
            elif x >= dw:
                x = max(0, dw - CLAMP_OVERLAP)
            if y + h < 0:
                y = min(0, CLAMP_OVERLAP - h)
            elif y > dh:
                y = max(0, dh - CLAMP_OVERLAP)
        return x, y

    def update_desktop_geometry(self, width: int, height: int) -> None:
        if self.desktop_geometry == (width, height):
            return  # no need to do anything
        self.desktop_geometry = (width, height)
        with xsync:
            geom = X11Window.getGeometry(self.corral_xid)
            assert geom
            x, y, w, h = geom[:4]
            nx, ny = self._clamp_to_desktop(x, y, w, h)
            if nx != x or ny != y:
                log("update_desktop_geometry(%i, %i) adjusting corral window to new location: %i,%i", width, height, nx, ny)
                X11Window.MoveWindow(self.corral_xid, nx, ny)

    def _read_initial_X11_properties(self) -> None:
        metalog("read_initial_X11_properties() window")

        # WARNING: have to handle _NET_WM_STATE before we look at WM_HINTS;
        # WM_HINTS assumes that our "state" property is already set.  This is
        # because there are four ways a window can get its urgency
        # ("attention-requested") bit set:
        #   1) _NET_WM_STATE_DEMANDS_ATTENTION in the _initial_ state hints
        #   2) setting the bit WM_HINTS, at _any_ time
        #   3) sending a request to the root window to add
        #      _NET_WM_STATE_DEMANDS_ATTENTION to their state hints
        #   4) if we (the wm) decide they should be and set it
        # To implement this, we generally track the urgency bit via
        # _NET_WM_STATE (since that is under our sole control during normal
        # operation).  Then (1) is accomplished through the normal rule that
        # initial states are read off from the client, and (2) is accomplished
        # by having WM_HINTS affect _NET_WM_STATE.  But this means that
        # WM_HINTS and _NET_WM_STATE handling become intertangled.

        def set_if_unset(propname: str, value):
            # the property may not be initialized yet,
            # if that's the case then calling get_property throws an exception:
            try:
                if self.get_property(propname) not in (None, ""):
                    return
            except TypeError:
                pass
            self._internal_set_property(propname, value)

        # "decorations" needs to be set before reading the X11 properties
        # because handle_wm_normal_hints_change reads it:
        set_if_unset("decorations", -1)
        super()._read_initial_X11_properties()
        net_wm_state = self.get_property("state")
        assert net_wm_state is not None, "_NET_WM_STATE should have been read already"
        geom = X11Window.getGeometry(self.xid)
        if not geom:
            raise Unmanageable(f"failed to get geometry for {self.xid:x}")
        # initial position and size, from the Window object,
        # but allow size hints to override it if specified
        x, y, w, h = geom[:4]

        def intpair(key: str, default: tuple[int, int]) -> tuple[int, int]:
            value = self.get(key)
            if value:
                return int(value[0]), int(value[1])
            return default
        ax, ay = intpair("requested-position", (x, y))
        aw, ah = intpair("requested-size", (w, h))
        geomlog("initial X11 position and size: requested(%s)=%s", geom, (ax, ay, aw, ah))
        set_if_unset("modal", "_NET_WM_STATE_MODAL" in net_wm_state)
        if x != ax or y != ay:
            self._internal_set_property("set-initial-position", True)
        self._internal_set_property("requested-position", (ax, ay))
        self._internal_set_property("requested-size", (aw, ah))
        self.update_children()

    def do_unmanaged(self, wm_exiting) -> None:
        log("unmanaging window: %s (%s - %s)", self, self.corral_xid, self.xid)
        cxid = self.corral_xid
        if cxid:
            self.corral_xid = 0
            remove_event_receiver(cxid, self)
            geom = None
            # use a new context, so we will XSync right here
            # and detect if the window is already gone:
            with xswallow:
                geom = X11Window.getGeometry(self.xid)
            if geom is not None:
                with xswallow:
                    if self.client_reparented:
                        X11Window.Reparent(self.xid, X11Window.get_root_xid(), 0, 0)
                    if self.saved_events != -1:
                        with xswallow:
                            X11Window.setEventMask(self.xid, self.saved_events)
            self.client_reparented = False
            # It is important to remove from our save set, even after
            # reparenting, because according to the X spec, windows that are
            # in our save set are always Mapped when we exit, *even if those
            # windows are no longer inferior to any of our windows!* (see
            # section 10. Connection Close).  This causes "ghost windows", see
            # bug #27:
            if self.in_save_set:
                with xswallow:
                    X11Window.XRemoveFromSaveSet(self.xid)
                self.in_save_set = False
            if geom:
                self.send_configure_notify()
            if wm_exiting:
                with xswallow:
                    if not X11Window.is_mapped(self.xid):
                        X11Window.MapWindow(self.xid)
            # it is now safe to destroy the corral window:
            with xsync:
                X11Window.DestroyWindow(cxid)
        super().do_unmanaged(wm_exiting)

    def send_configure_notify(self) -> None:
        with xlog:
            return X11Window.sendConfigureNotify(self.xid)

    #########################################
    # Actions specific to WindowModel
    #########################################

    def raise_window(self) -> None:
        X11Window.XRaiseWindow(self.corral_xid)
        X11Window.XRaiseWindow(self.xid)

    def unmap(self) -> None:
        with xsync:
            if X11Window.is_mapped(self.xid):
                self.last_unmap_serial = X11Window.Unmap(self.xid)
                log("client window %#x unmapped, serial=%#x", self.xid, self.last_unmap_serial)
                self._internal_set_property("shown", False)

    def map(self) -> None:
        with xsync:
            if not X11Window.is_mapped(self.xid):
                X11Window.MapWindow(self.xid)
                log("client window %#x mapped", self.xid)

    #########################################
    # X11 Events
    #########################################

    def do_x11_property_notify_event(self, event) -> None:
        if event.delivered_to == self.corral_xid:
            return
        super().do_x11_property_notify_event(event)

    # noinspection PyMethodMayBeStatic
    def do_x11_child_map_request_event(self, event) -> None:
        # If we get a MapRequest then it might mean that someone tried to map
        # this window multiple times in quick succession, before we actually
        # mapped it (so that several MapRequests ended up queued up; FSF Emacs
        # 22.1.50.1 does this, at least).  It alternatively might mean that
        # the client is naughty and tried to map their window which is
        # currently not displayed.  In either case, we should just ignore the
        # request.
        log("do_x11_child_map_request_event(%s)", event)

    def do_x11_unmap_event(self, event) -> None:
        log(f"do_x11_unmap_event({event}) corral_xid={self.corral_xid:x}")
        if not self.corral_xid or event.delivered_to == self.corral_xid:
            return
        if event.window != self.xid:
            log.warn(f"Warning: unexpected unmap event for window {event.window:x}")
            return
        # The client window got unmapped.  The question is, though, was that
        # because it was withdrawn/destroyed, or was it because we unmapped it
        # going into `IconicState`?
        #
        # Also, if we receive a *synthetic* UnmapNotify event, that always
        # means that the client has withdrawn the window (even if it was not
        # mapped in the first place) -- ICCCM section 4.1.4.
        log("do_x11_unmap_event(%s) client window unmapped, last_unmap_serial=%#x", event, self.last_unmap_serial)
        if event.send_event or self.serial_after_last_unmap(event.serial):
            self.unmanage()

    def do_x11_destroy_event(self, event) -> None:
        log(f"do_x11_destroy_event({event}) corral_xid={self.corral_xid:x}")
        if not self.corral_xid or event.delivered_to == self.corral_xid:
            return
        assert event.window == self.xid
        super().do_x11_destroy_event(event)

    #########################################
    # Hooks for WM
    #########################################

    def show(self) -> None:
        self.map()
        self._internal_set_property("shown", True)
        if self.get_property("iconic"):
            self.set_property("iconic", False)
        self._update_client_geometry()
        with xlog:
            if not X11Window.is_mapped(self.corral_xid):
                X11Window.MapWindow(self.corral_xid)

    def hide(self) -> None:
        self._internal_set_property("shown", False)
        with xlog:
            if X11Window.is_mapped(self.corral_xid):
                X11Window.Unmap(self.corral_xid)
            if X11Window.getParent(self.corral_xid) != self.parking_window_xid:
                X11Window.Reparent(self.corral_xid, self.parking_window_xid, 0, 0)
            X11Window.sendConfigureNotify(self.xid)

    def _update_client_geometry(self) -> None:
        """ figure out where we're supposed to get the window geometry from,
            and call do_update_client_geometry which will send a Configure and Notify
        """
        geometry = self.get_property("client-geometry")
        if geometry is not None:
            geomlog("_update_client_geometry: using client-geometry=%s", geometry)
            x, y, w, h = geometry
            self._internal_set_property("set-initial-position", True)
            self._internal_set_property("requested-position", (x, y))
            self._internal_set_property("requested-size", (w, h))
        elif not self._setup_done:
            # try to honour initial size and position requests during setup:
            with xsync:
                geom = X11Window.getGeometry(self.xid)
                if not geom:
                    raise XError(f"window {self.xid:x} disappeared")
                x, y, w, h = geom[:4]
            x, y = self.get_property("requested-position") or (x, y)
            w, h = self.get_property("requested-size") or (w, h)
            geometry = x, y, w, h
            geomlog("_update_client_geometry: using initial geometry=%s", geometry)
        else:
            geometry = self.get_property("geometry")
            geomlog("_update_client_geometry: using current geometry=%s", geometry)
        self._do_update_client_geometry(*geometry)

    def _do_update_client_geometry(self, x, y, allocated_w, allocated_h) -> None:
        geomlog("_do_update_client_geometry(%s)", (x, y, allocated_w, allocated_h))
        hints = self.get_property("size-hints")
        w, h = self.calc_constrained_size(allocated_w, allocated_h, hints)
        geomlog("_do_update_client_geometry: size(%s)=%ix%i", hints, w, h)
        with xlog:
            if self.corral_xid:
                geom = X11Window.getGeometry(self.corral_xid)
                assert geom
                cx, cy, cw, ch = geom[:4]
                if cx != x or cy != y or cw != w or ch != h:
                    X11Window.MoveResizeWindow(self.corral_xid, x, y, w, h)
                    X11Window.configure(self.xid, 0, 0, w, h)
                    X11Window.sendConfigureNotify(self.xid)
                else:
                    X11Window.sendConfigureNotify(self.xid)
            else:
                # corral window hasn't been created yet
                X11Window.configure(self.xid, x, y, w, h)
                X11Window.sendConfigureNotify(self.xid)
            self._updateprop("geometry", (x, y, w, h))

    def do_x11_configure_event(self, event) -> None:
        geomlog("WindowModel.do_x11_configure_event(%s) corral=%#x, client=%#x, managed=%s",
                event, self.corral_xid, self.xid, self._managed)
        if not self._managed or not self.corral_xid or not self.xid:
            return
        if event.window == self.corral_xid:
            # we only care about events on the client window
            geomlog(f"ignored configure event on the corral window {self.corral_xid:x}")
            return
        if event.window != self.xid:
            # we only care about events on the client window
            geomlog(f"ignored configure event on window {event.window:x} (not the client window)")
            return
        try:
            # workaround applications whose windows disappear from underneath us:
            with xsync:
                if not X11Window.is_mapped(self.corral_xid):
                    geomlog(f"WindowModel.do_x11_configure_event: corral window {self.corral_xid:x} is not visible")
                    return
                if not X11Window.is_mapped(self.xid):
                    geomlog(f"WindowModel.do_x11_configure_event: client window {self.xid:x} is not visible")
                    return
                # event.border_width unused
                self.configure_geometry(event.x, event.y, event.width, event.height)
                self.update_children()
        except XError as e:
            geomlog("do_x11_configure_event(%s)", event, exc_info=True)
            geomlog.warn("Warning: failed to resize corral window %#x", self.corral_xid)
            geomlog.warn(" %s", e)

    def update_children(self) -> None:
        geom = X11Window.getGeometry(self.xid)
        if not geom:
            return
        ww, wh = geom[2:4]
        children = []
        for xid in X11Window.get_children(self.xid):
            if X11Window.is_inputonly(xid):
                continue
            geom = X11Window.getGeometry(xid)
            if not geom:
                continue
            if geom[2] == geom[3] == 1:
                # skip 1x1 windows, as those are usually just event windows
                continue
            if geom[0] == geom[1] == 0 and geom[2] == ww and geom[3] == wh:
                # exact same geometry as the window itself
                continue
            # record xid and geometry:
            children.append([xid] + list(geom))
        self._internal_set_property("children", children)

    def configure_geometry(self, x: int, y: int, w: int, h: int) -> bool:
        # the client window may have been resized or moved (generally programmatically)
        # so we may need to update the corral_window to match
        geomlog("configure_geometry%s", (x, y, w, h))
        cow, coh = X11Window.getGeometry(self.corral_xid)[2:4]
        # size changes (and position if any):
        hints = self.get_property("size-hints")
        w, h = self.calc_constrained_size(w, h, hints)
        cx, cy, cw, ch = self.get_property("geometry")
        resized = cow != w or coh != h
        moved = x != 0 or y != 0
        geomlog("configure_geometry%s hints=%s, constrained size=%s, corral=%s, geometry=%s, resized=%s, moved=%s",
                (x, y, w, h), hints, (w, h), (cow, coh), (cx, cy, cw, ch), resized, moved)
        if not (moved or resized):
            return False
        if moved:
            self._internal_set_property("set-initial-position", True)
            self._internal_set_property("requested-position", (x, y))
        if resized:
            self._internal_set_property("requested-size", (w, h))
        if moved and resized:
            X11Window.MoveResizeWindow(self.corral_xid, x, y, w, h)
            X11Window.MoveWindow(self.xid, 0, 0)
        elif moved:
            # ICCCM 4.2.3
            X11Window.MoveWindow(self.corral_xid, x, y)
            X11Window.MoveWindow(self.xid, 0, 0)
            self.send_configure_notify()
        elif resized:
            X11Window.ResizeWindow(self.corral_xid, w, h)
            x = cx
            y = cy
        return self._updateprop("geometry", (x, y, w, h))

    def do_x11_child_configure_request_event(self, event) -> None:
        hints = self.get_property("size-hints")
        geomlog("do_x11_child_configure_request_event(%s) client=%#x, corral=%#x, value_mask=%s, size-hints=%s",
                event, self.xid, self.corral_xid, configure_bits(event.value_mask), hints)
        if event.value_mask & CWStackMode:
            geomlog(" restack above=%s, detail=%s", event.above, event.detail)
        # Also potentially update our record of what the app has requested:
        ox, oy, ow, oh = self.get_property("geometry")
        x, y, w, h = ox, oy, ow, oh
        rx, ry = self.get_property("requested-position")
        if event.value_mask & CWX:
            x = event.x
            rx = x
        if event.value_mask & CWY:
            y = event.y
            ry = y
        if event.value_mask & CWX or event.value_mask & CWY:
            self._internal_set_property("set-initial-position", True)
            self._internal_set_property("requested-position", (rx, ry))

        rw, rh = self.get_property("requested-size")
        if event.value_mask & CWWidth:
            w = event.width
            rw = w
        if event.value_mask & CWHeight:
            h = event.height
            rh = h
        if event.value_mask & CWWidth or event.value_mask & CWHeight:
            self._internal_set_property("requested-size", (rw, rh))

        if event.value_mask & CWStackMode:
            self.emit("restack", event.detail, event.above)

        if VALIDATE_CONFIGURE_REQUEST:
            w, h = self.calc_constrained_size(w, h, hints)

        # update the geometry now, as another request may come in
        # before we've had a chance to process the ConfigureNotify that the code below will generate
        if not self._updateprop("geometry", (x, y, w, h)):
            # As per ICCCM 4.1.5, even if we ignore the request
            # send back a synthetic ConfigureNotify telling the client that nothing has happened.
            geomlog("geometry unchanged: %s", (x, y, w, h))
            with xlog:
                X11Window.sendConfigureNotify(self.xid)
            return

        geomlog("do_x11_child_configure_request_event updated requested geometry from %s to %s",
                (ox, oy, ow, oh), (x, y, w, h))
        mask = 0
        if ox != x:
            mask |= CWX
        if oy != y:
            mask |= CWY
        if ow != w:
            mask |= CWWidth
        if oh != h:
            mask |= CWHeight
        with xlog:
            X11Window.configure(self.xid, x, y, w, h, mask)
        # FIXME: consider handling attempts to change stacking order here.
        # (In particular, I believe that a request to jump to the top is
        # meaningful and should perhaps even be respected.)

    def process_client_message_event(self, event) -> bool:
        if event.message_type == "_NET_MOVERESIZE_WINDOW":
            # TODO: honour gravity, show source indication
            with xsync:
                geom = X11Window.getGeometry(self.corral_xid)
                x, y, w, h, _ = geom
                if event.data[0] & 0x100:
                    x = event.data[1]
                if event.data[0] & 0x200:
                    y = event.data[2]
                if event.data[0] & 0x400:
                    w = event.data[3]
                if event.data[0] & 0x800:
                    h = event.data[4]
                if (event.data[0] & 0x100) or (event.data[0] & 0x200):
                    self._internal_set_property("set-initial-position", True)
                    self._internal_set_property("requested-position", (x, y))
                # honour hints:
                hints = self.get_property("size-hints")
                w, h = self.calc_constrained_size(w, h, hints)
                geomlog("_NET_MOVERESIZE_WINDOW on %s (data=%s, current geometry=%s, new geometry=%s)",
                        self, event.data, geom, (x, y, w, h))
                X11Window.configure(self.xid, x, y, w, h)
                X11Window.sendConfigureNotify(self.xid)
            return True
        return super().process_client_message_event(event)

    def calc_constrained_size(self, w: int, h: int, hints: dict) -> tuple[int, int]:
        mhints = typedict(hints)
        cw, ch = calc_constrained_size(w, h, mhints)
        geomlog("calc_constrained_size%s=%s (size_constraints=%s)", (w, h, mhints), (cw, ch), self.size_constraints)
        return cw, ch

    def update_size_constraints(self, minw: int = 0, minh: int = 0,
                                maxw: int = MAX_WINDOW_SIZE, maxh: int = MAX_WINDOW_SIZE) -> None:
        if self.size_constraints == (minw, minh, maxw, maxh):
            geomlog("update_size_constraints%s unchanged", (minw, minh, maxw, maxh))
            return  # no need to do anything
        ominw, ominh, omaxw, omaxh = self.size_constraints
        self.size_constraints = minw, minh, maxw, maxh
        if minw <= ominw and minh <= ominh and maxw >= omaxw and maxh >= omaxh:
            geomlog("update_size_constraints%s less restrictive, no need to recalculate", (minw, minh, maxw, maxh))
            return
        geomlog("update_size_constraints%s recalculating client geometry", (minw, minh, maxw, maxh))
        if self.get_property("shown"):
            self._update_client_geometry()

    #########################################
    # X11 properties synced to Python objects
    #########################################

    def _handle_icon_title_change(self) -> None:
        icon_name = self.prop_get("_NET_WM_ICON_NAME", "utf8", True)
        iconlog("_NET_WM_ICON_NAME=%s", icon_name)
        if not icon_name:
            icon_name = self.prop_get("WM_ICON_NAME", "latin1", True)
            iconlog("WM_ICON_NAME=%s", icon_name)
        self._updateprop("icon-title", sanestr(str(icon_name) or ""))

    def _handle_motif_wm_hints_change(self) -> None:
        motif_hints = self.prop_get("_MOTIF_WM_HINTS", "motif-hints", ignore_errors=False, raise_xerrors=True)
        metalog("_MOTIF_WM_HINTS=%s", motif_hints)
        if motif_hints:
            if motif_hints.flags & (2 ** MotifWMHints.DECORATIONS_BIT):
                if self._updateprop("decorations", motif_hints.decorations):
                    # we may need to clamp the window size:
                    self._handle_wm_normal_hints_change()
            if motif_hints.flags & (2 ** MotifWMHints.INPUT_MODE_BIT):
                self._updateprop("modal", int(motif_hints.input_mode))

    def _handle_wm_normal_hints_change(self) -> None:
        with xswallow:
            size_hints = X11Window.getSizeHints(self.xid)
        metalog("WM_NORMAL_HINTS=%s", size_hints)
        # getSizeHints exports fields using their X11 names as defined in the "XSizeHints" structure,
        # but we use a different naming (for historical reason and backwards compatibility)
        # so rename the fields:
        hints = {}
        if size_hints:
            TRANSLATED_NAMES: dict[str, str] = {
                "position": "position",
                "size": "size",
                "base_size": "base-size",
                "resize_inc": "increment",
                "win_gravity": "gravity",
                "min_aspect_ratio": "minimum-aspect-ratio",
                "max_aspect_ratio": "maximum-aspect-ratio",
            }
            for k, v in size_hints.items():
                trans_name = TRANSLATED_NAMES.get(k)
                if trans_name:
                    hints[trans_name] = v
        # handle min-size and max-size,
        # applying our size constraints if we have any:
        mhints = typedict(size_hints or {})
        hminw, hminh = mhints.inttupleget("min_size", (0, 0), 2, 2)
        hmaxw, hmaxh = mhints.inttupleget("max_size", (MAX_WINDOW_SIZE, MAX_WINDOW_SIZE), 2, 2)
        # noinspection PyTypeChecker
        d = int(self.get("decorations", -1))
        decorated = d == -1 or any(
            (d & 2 ** b) for b in (
                MotifWMHints.ALL_BIT,
                MotifWMHints.TITLE_BIT,
                MotifWMHints.MINIMIZE_BIT,
                MotifWMHints.MAXIMIZE_BIT,
            )
        )
        cminw, cminh, cmaxw, cmaxh = self.size_constraints
        if decorated:
            # min-size only applies to decorated windows
            if cminw > 0 and cminw > hminw:
                hminw = cminw
            if cminh > 0 and cminh > hminh:
                hminh = cminh
        # max-size applies to all windows:
        if 0 < cmaxw < hmaxw:
            hmaxw = cmaxw
        if 0 < cmaxh < hmaxh:
            hmaxh = cmaxh
        # if the values mean something, expose them:
        if hminw > 0 or hminh > 0:
            hints["minimum-size"] = hminw, hminh
        if hmaxw < MAX_WINDOW_SIZE or hmaxh < MAX_WINDOW_SIZE:
            hints["maximum-size"] = hmaxw, hmaxh
        sanitize_size_hints(hints)

        # size and position are stored as properties, not as hints:
        pos = hints.pop("position", None)
        if pos is not None and SIZEHINTS_POSITION:
            self._internal_set_property("requested-position", pos)
            self._internal_set_property("set-initial-position", True)
        size = hints.pop("size", None)
        if size is not None and SIZEHINTS_SIZE:
            self._internal_set_property("requested-size", size)

        # Don't send out notify and ConfigureNotify events when this property
        # gets no-op updated -- some apps like FSF Emacs 21 like to update
        # their properties every time they see a ConfigureNotify, and this
        # reduces the chance for us to get caught in loops:
        if self._updateprop("size-hints", hints):
            metalog("updated: size-hints=%s", hints)
            self._update_client_geometry()

    def _handle_net_wm_icon_change(self) -> None:
        iconlog("_NET_WM_ICON changed on %#x, re-reading", self.xid)
        icons = self.prop_get("_NET_WM_ICON", "icons")
        self._internal_set_property("icons", icons)

    _x11_property_handlers = dict(BaseWindowModel._x11_property_handlers)
    _x11_property_handlers |= {
        "WM_ICON_NAME": _handle_icon_title_change,
        "_NET_WM_ICON_NAME": _handle_icon_title_change,
        "_MOTIF_WM_HINTS": _handle_motif_wm_hints_change,
        "WM_NORMAL_HINTS": _handle_wm_normal_hints_change,
        "_NET_WM_ICON": _handle_net_wm_icon_change,
    }

    # noinspection PyMethodMayBeStatic
    def get_default_window_icon(self, _size: int = 48):
        return None

    def get_wm_state(self, prop: str) -> bool:
        state_names = self._state_properties.get(prop)
        assert state_names, f"invalid window state {prop}"
        log("get_wm_state(%s) state_names=%s", prop, state_names)
        # this is a virtual property for _NET_WM_STATE:
        # return True if any is set (only relevant for maximized)
        return any(self._state_isset(x) for x in state_names)

    ################################
    # Focus handling:
    ################################

    def give_client_focus(self) -> None:
        """The focus manager has decided that our client should receive X
        focus.  See world_window.py for details."""
        log("give_client_focus() corral_xid=%s", self.corral_xid)
        if self.corral_xid:
            with xlog:
                self.do_give_client_focus()

    def do_give_client_focus(self) -> None:
        protocols = self.get_property("protocols")
        focuslog("Giving focus to %#x, input_field=%s, FORCE_XSETINPUTFOCUS=%s, protocols=%s",
                 self.xid, self._input_field, FORCE_XSETINPUTFOCUS, protocols)
        # Have to fetch the time, not just use CurrentTime, both because ICCCM
        # says that WM_TAKE_FOCUS must use a real time and because there are
        # genuine race conditions here (e.g. suppose the client does not
        # actually get around to requesting the focus until after we have
        # already changed our mind and decided to give it to someone else).
        now = X11Window.get_server_time(self.corral_xid)
        # ICCCM 4.1.7 *claims* to describe how we are supposed to give focus
        # to a window, but it is completely opaque.  From reading the
        # metacity, kwin, gtk+, and qt code, it appears that the actual rules
        # for giving focus are:
        #   -- the WM_HINTS input field determines whether the WM should call
        #      XSetInputFocus
        #   -- independently, the WM_TAKE_FOCUS protocol determines whether
        #      the WM should send a WM_TAKE_FOCUS ClientMessage.
        # If both are set, both methods MUST be used together. For example,
        # GTK+ apps respect WM_TAKE_FOCUS alone, but I'm not sure they handle
        # XSetInputFocus well, while Qt apps ignore (!!!) WM_TAKE_FOCUS
        # (unless they have a modal window), and just expect to get focus from
        # the WM's XSetInputFocus.
        if bool(self._input_field) or FORCE_XSETINPUTFOCUS:
            focuslog("... using XSetInputFocus")
            X11Window.XSetInputFocus(self.xid, now)
        if "WM_TAKE_FOCUS" in protocols:
            focuslog("... using WM_TAKE_FOCUS")
            send_wm_take_focus(self.xid, now)
        self.set_active()


GObject.type_register(WindowModel)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.util.objects import AdHocStruct
from unit.server.mixins.servermixintest_util import ServerMixinTest


class ServerMixinsTest(ServerMixinTest):

    def test_remotelogging(self):
        from xpra.server.mixins.logging import LoggingServer
        messages = []
        def newlogfn(*args):
            messages.append(args)
        def _LoggingServer():
            ls = LoggingServer()
            ls.do_log = newlogfn
            return ls
        opts = AdHocStruct()
        opts.remote_logging = "on"
        level = 20
        msg = "foo"
        packet = ["logging", level, msg]
        self._test_mixin_class(_LoggingServer, opts)
        self.handle_packet(packet)
        assert len(messages)==1

def main():
    unittest.main()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

#!python
#cython: boundscheck=False, wraparound=False

from typing import Any, Dict

from xpra.log import Logger
log = Logger("util", "auth")

from ctypes import addressof, create_string_buffer, sizeof

from libc.stdint cimport uintptr_t   # pylint: disable=syntax-error
from libc.string cimport memset


cdef extern from "Python.h":
    int PyObject_GetBuffer(object obj, Py_buffer *view, int flags)
    void PyBuffer_Release(Py_buffer *view)
    int PyBUF_ANY_CONTIGUOUS

cdef extern from "errno.h" nogil:
    int errno

cdef extern from "string.h":
    char *strdup(const char *s1)

cdef extern from "stdlib.h":
    void *calloc(size_t nitems, size_t size)

cdef extern from "security/pam_misc.h":
    ctypedef struct pam_handle_t:
        pass
    void misc_conv(int num_msg, const pam_message **msgm, pam_response **response, void *appdata_ptr)

cdef extern from "security/pam_appl.h":
    struct pam_conv:
        void *conv
        #int (*conv)(int num_msg, const pam_message **msg, pam_response **resp, void *appdata_ptr)
        void *appdata_ptr
    struct pam_message:
        int msg_style
        const char *msg
    struct pam_response:
        char *resp
        int resp_retcode
    struct pam_xauth_data:
        int namelen
        char *name
        int datalen
        char *data

    const char *pam_strerror(pam_handle_t *pamh, int errnum)
    int pam_start(const char *service_name, const char *user, const pam_conv *pam_conversation, pam_handle_t **pamh)
    int pam_open_session(pam_handle_t *pamh, int flags)
    int pam_close_session(pam_handle_t *pamh, int flags)
    int pam_end(pam_handle_t *pamh, int pam_status)
    int pam_putenv(pam_handle_t *pamh, const char *name_value)
    int pam_set_item(pam_handle_t *pamh, int item_type, const void *item)
    char **pam_getenvlist(pam_handle_t *pamh)

    int pam_authenticate(pam_handle_t *pamh, int flags)
    int pam_acct_mgmt(pam_handle_t *pamh, int flags)

    int PAM_SERVICE             # The service name
    int PAM_USER                # The user name
    int PAM_TTY                 # The tty name
    int PAM_RHOST               # The remote host name
    int PAM_CONV                # The pam_conv structure
    int PAM_AUTHTOK             # The authentication token (password)
    int PAM_OLDAUTHTOK          # The old authentication token
    int PAM_RUSER               # The remote user name
    int PAM_USER_PROMPT         # the prompt for getting a username
    int PAM_FAIL_DELAY          # app supplied function to override failure
    int PAM_XDISPLAY            # X display name
    int PAM_XAUTHDATA           # X server authentication data
    int PAM_AUTHTOK_TYPE        # The type for pam_get_authtok

    int PAM_PROMPT_ECHO_ON
    int PAM_PROMPT_ECHO_OFF
    int PAM_PROMPT_TEXT_INFO

    int PAM_SILENT

    int PAM_ABORT               #The application should exit immediately after calling pam_end(3) first
    int PAM_AUTH_ERR            #The user was not authenticated
    int PAM_CRED_INSUFFICIENT   #For some reason the application does not have sufficient credentials to authenticate the user
    int PAM_AUTHINFO_UNAVAIL    #The modules were not able to access the authentication information. This might be due to a network or hardware failure etc
    int PAM_MAXTRIES            #One or more of the authentication modules has reached its limit of tries authenticating the user. Do not try again
    int PAM_SUCCESS             #The user was successfully authenticated
    int PAM_USER_UNKNOWN        #User unknown to authentication service

PAM_ERR_STR: Dict[int, str] = {
    PAM_ABORT               : "ABORT",
    PAM_AUTH_ERR            : "AUTH_ERR",
    PAM_CRED_INSUFFICIENT   : "CRED_INSUFFICIENT",
    PAM_AUTHINFO_UNAVAIL    : "AUTHINFO_UNAVAIL",
    PAM_MAXTRIES            : "MAXTRIES",
    PAM_SUCCESS             : "SUCCESS",
    PAM_USER_UNKNOWN        : "USER_UNKNOWN",
}

PAM_ITEMS: Dict[str, int] = {
    "SERVICE"       : PAM_SERVICE,
    "USER"          : PAM_USER,
    "TTY"           : PAM_TTY,
    "RHOST"         : PAM_RHOST,
    "CONV"          : PAM_CONV,
    "AUTHTOK"       : PAM_AUTHTOK,
    "OLDAUTHTOK"    : PAM_OLDAUTHTOK,
    "RUSER"         : PAM_RUSER,
    "USER_PROMPT"   : PAM_USER_PROMPT,
    "FAIL_DELAY"    : PAM_FAIL_DELAY,
    "XDISPLAY"      : PAM_XDISPLAY,
    "XAUTHDATA"     : PAM_XAUTHDATA,
    "AUTHTOK_TYPE"  : PAM_AUTHTOK_TYPE,
}


cdef inline bytes b(value: str):
    return value.encode("latin1")


cdef int password_conv(int n_msg, const pam_message **msg, pam_response **resp, void *appdata_ptr):
    if appdata_ptr==NULL:
        return 1
    cdef pam_response* response = <pam_response*> calloc(n_msg, sizeof(pam_response))
    if response==NULL:
        return 1
    cdef const char *password = <const char*> appdata_ptr  # @UndefinedVariable
    resp[0] = response
    for i in range(n_msg):
        if msg[i].msg_style == PAM_PROMPT_ECHO_OFF:
            response[i].resp = strdup(password)
            response[i].resp_retcode = 0
    return 0


cdef class pam_session:

    cdef pam_handle_t *pam_handle
    cdef object service_name
    cdef object username
    cdef object password

    def __init__(self, username: str, password="", service_name="xpra"):
        # convert to bytes for cython calls:
        self.service_name = b(service_name)
        self.username = b(username)
        self.password = b(password)
        self.pam_handle = NULL

    def __repr__(self):
        return "pam_session(%#x)" % (<uintptr_t> self.pam_handle)

    def pam_strerror(self, int r) -> str:
        berr = pam_strerror(self.pam_handle, r)
        try:
            return berr.decode("latin1")
        except UnicodeError:
            return str(berr)

    def start(self, password=False) -> bool:
        cdef pam_conv conv
        cdef Py_buffer view
        memset(&view, 0, sizeof(Py_buffer))

        if self.pam_handle != NULL:
            log.error("Error: cannot open the pam session more than once!")
            return False

        if password:
            conv.conv = <void *> &password_conv
            if not self.password:
                raise RuntimeError("no password to use for pam_start")
            if PyObject_GetBuffer(self.password, &view, PyBUF_ANY_CONTIGUOUS):
                raise RuntimeError("failed to read password data")
            conv.appdata_ptr = view.buf
        else:
            conv.conv = <void *> misc_conv
            conv.appdata_ptr = NULL
        cdef int r = 0
        try:
            r = pam_start(self.service_name, self.username, &conv, &self.pam_handle)
        finally:
            if password and view.buf != NULL:
                PyBuffer_Release(&view)
        log("pam_start: %s", PAM_ERR_STR.get(r, r))
        if r != PAM_SUCCESS:
            self.pam_handle = NULL
            log.error("Error: pam_start failed: %s", PAM_ERR_STR.get(r, r))
            log.error(" %r", self.pam_strerror(r))
            return False
        return True

    def set_env(self, env : Dict[str,Any]) -> None:
        assert self.pam_handle!=NULL
        cdef int r
        for k,v in env.items():
            name_value = "%s=%s\0" % (k, v)
            r = pam_putenv(self.pam_handle, b(name_value))
            if r != PAM_SUCCESS:
                log.error("Error %i: failed to add '%s' to pam environment", r, name_value)
                log.error(" %r", self.pam_strerror(r))
            else:
                log("pam_putenv: %s", name_value)

    def get_envlist(self) -> Dict[str,str]:
        assert self.pam_handle!=NULL
        cdef char **envlist = pam_getenvlist(self.pam_handle)
        log("pam_getenvlist: %#x", <uintptr_t> envlist)
        cdef unsigned int i
        env = {}
        if envlist!=NULL:
            i = 0
            while envlist[i]!=NULL:
                bstr = envlist[i]
                s = bstr.decode("latin1")
                parts = s.split("=", 1)
                if len(parts)==2:
                    env[parts[0]] = parts[1]
                i += 1
        log("get_envlist()=%s", env)
        return env

    def set_items(self, items : Dict[str,Any]) -> None:
        cdef const void* item
        cdef pam_xauth_data xauth_data
        cdef int r
        assert self.pam_handle!=NULL
        for k,v in items.items():
            v = b(v)
            item_type = PAM_ITEMS.get(k.upper())
            log("item_type(%s)=%s", k, item_type)
            if item_type is None or item_type in (PAM_CONV, PAM_FAIL_DELAY):
                log.error("Error: invalid pam item '%s'", k)
                continue
            elif item_type==PAM_XAUTHDATA:
                method = b"MIT-MAGIC-COOKIE-1\0"
                xauth_data.namelen = len(b"MIT-MAGIC-COOKIE-1")
                xauth_data.name = method
                s = v+b"\0"
                xauth_data.datalen = len(v)
                xauth_data.data = s
                item = <const void *> &xauth_data
            else:
                s = create_string_buffer(v)
                l = addressof(s)
                item = <const void *> l
            r = pam_set_item(self.pam_handle, item_type, item)
            if r != PAM_SUCCESS:
                log.error("Error %i: failed to set pam item '%s' to '%s'", r, k, v)
                log.error(" %r", self.pam_strerror(r))
            else:
                log("pam_set_item: %s=%s", k, v)

    def open(self) -> bool:         #@ReservedAssignment
        assert self.pam_handle!=NULL
        cdef int r = pam_open_session(self.pam_handle, 0)
        log("pam_open_session: %s", PAM_ERR_STR.get(r, r))
        if r != PAM_SUCCESS:
            self.pam_handle = NULL
            log.error("Error: pam_open_session failed: %s", PAM_ERR_STR.get(r, r))
            log.error(" %r", self.pam_strerror(r))
            return False
        return True

    def authenticate(self) -> bool:
        log("authenticate()")
        assert self.pam_handle!=NULL
        cdef int r = pam_authenticate(self.pam_handle, PAM_SILENT)
        log("pam_authenticate: %s", PAM_ERR_STR.get(r, r))
        if r != PAM_SUCCESS:
            log.warn("Warning: pam authentication failed: %s", PAM_ERR_STR.get(r, r))
            log.warn(" %r", self.pam_strerror(r))
            return False
        return True

    def check_account(self) -> bool:
        assert self.pam_handle!=NULL
        cdef int r = pam_acct_mgmt(self.pam_handle, 0)
        if r != PAM_SUCCESS:
            self.pam_handle = NULL
            log.error("Error: pam_acct_mgmt failed: %s", PAM_ERR_STR.get(r, r))
            log.error(" %r", self.pam_strerror(r))
            return False
        return True

    def close(self) -> bool:
        if self.pam_handle==NULL:
            log.error("Error: no pam session to close!")
            return False

        cdef int r = pam_close_session(self.pam_handle, 0)
        log("pam_close_session: %s", PAM_ERR_STR.get(r, r))
        if r != PAM_SUCCESS:
            self.pam_handle = NULL
            log.error("Error: failed to close the pam session: %s", PAM_ERR_STR.get(r, r))
            log.error(" %r", self.pam_strerror(r))
            return False

        r = pam_end(self.pam_handle, r)
        log("pam_end: %s", PAM_ERR_STR.get(r, r))
        self.pam_handle = NULL
        if r != PAM_SUCCESS:
            log.error("Error: pam_end '%s'", PAM_ERR_STR.get(r, r))
            log.error(" %r", self.pam_strerror(r))
            return False
        return True



# This file is part of Xpra.
# Copyright (C) 2021 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
from importlib import import_module
from typing import Any, TypeAlias

from xpra.util.parsing import parse_simple_dict
from xpra.os_util import WIN32
from xpra.scripts.config import InitException
from xpra.log import Logger

log = Logger("auth")


AuthDef: TypeAlias = tuple[str, Any, type, dict]


def get_auth_module(auth_str: str, cwd=os.getcwd(), **auth_options) -> AuthDef:
    log("get_auth_module(%s, {..})", auth_str)
    # separate options from the auth module name
    # either with ":" or "," as separator
    scpos = auth_str.find(":")
    cpos = auth_str.find(",")
    if cpos < 0 or scpos < cpos:
        parts = auth_str.split(":", 1)
    else:
        parts = auth_str.split(",", 1)
    auth = parts[0]
    if auth.endswith("base") or auth.endswith("helper"):
        raise ValueError(f"invalid authentication module name {auth}")
    if len(parts) > 1:
        auth_options.update(parse_simple_dict(parts[1]))
    auth_options["exec_cwd"] = cwd
    try:
        if auth == "sys":
            # resolve virtual "sys" auth:
            if WIN32:
                auth_modname = "win32"
            else:
                auth_modname = "pam"
            log("will try to use sys auth module '%s' for %s", auth, sys.platform)
        else:
            auth_modname = auth.replace("-", "_")
        auth_mod_name = "xpra.auth." + auth_modname
        log("auth module name for '%s': '%s'", auth, auth_mod_name)
        auth_module = import_module(auth_mod_name)
        if not hasattr(auth_module, "Authenticator"):
            raise ImportError(f"no 'Authenticator' class in {auth_mod_name!r}")
    except ImportError as e:
        log("cannot load %s auth for %r", auth, auth_str, exc_info=True)
        raise InitException(f"cannot load authentication module '{auth}' for {auth_str!r}: {e}") from None
    log("auth module for '%s': %s", auth, auth_module)
    try:
        auth_class = auth_module.Authenticator
        auth_class.auth_name = auth.lower()
        return auth, auth_module, auth_class, auth_options
    except Exception as e:
        log("cannot access authenticator class", exc_info=True)
        raise InitException(f"authentication setup error in {auth_module}: {e}") from None



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from xpra.os_util import gi_import
from xpra.util.objects import typedict
from xpra.util.str_fn import csv
from xpra.util.env import envint, envbool
from xpra.net.file_transfer import FileTransferHandler
from xpra.client.base.stub_client_mixin import StubClientMixin
from xpra.util.thread import start_thread
from xpra.log import Logger

GLib = gi_import("GLib")

printlog = Logger("printing")
filelog = Logger("file")

DELETE_PRINTER_FILE = envbool("XPRA_DELETE_PRINTER_FILE", True)
SKIP_STOPPED_PRINTERS = envbool("XPRA_SKIP_STOPPED_PRINTERS", True)
INIT_PRINTING_DELAY = envint("XPRA_INIT_PRINTING_DELAY", 2)


class FilePrintMixin(StubClientMixin, FileTransferHandler):

    def __init__(self):
        StubClientMixin.__init__(self)
        FileTransferHandler.__init__(self)
        self.printer_attributes = []
        self.send_printers_timer: int = 0
        self.exported_printers = None
        self.remote_request_file: bool = False

    def init(self, opts) -> None:
        # printing and file transfer:
        FileTransferHandler.init_opts(self, opts)

    def init_authenticated_packet_handlers(self) -> None:
        self.add_packets(
            "open-url", "send-file",
            "send-data-request", "send-data-response",
            "ack-file-chunk", "send-file-chunk",
        )

    def get_caps(self) -> dict[str, Any]:
        return self.get_file_transfer_features()

    def cleanup(self) -> None:
        # we must clean printing before FileTransferHandler, which turns the printing flag off!
        self.cleanup_printing()
        FileTransferHandler.cleanup(self)

    def parse_server_capabilities(self, c: typedict) -> bool:
        self.parse_printing_capabilities(c)
        self.parse_file_transfer_caps(c)
        self.remote_request_file = c.boolget("request-file", False)
        return True

    def parse_printing_capabilities(self, caps: typedict) -> None:
        printlog("parse_printing_capabilities() client printing support=%s", self.printing)
        if self.printing:
            server_printing = caps.boolget("printing")
            printlog("parse_printing_capabilities() server printing support=%s", server_printing)
            if server_printing:
                self.printer_attributes = caps.strtupleget("printer.attributes",
                                                           ("printer-info", "device-uri"))
                GLib.timeout_add(INIT_PRINTING_DELAY * 1000, self.init_printing)

    def init_printing(self) -> None:
        try:
            from xpra.platform.printing import init_printing  # pylint: disable=import-outside-toplevel
            printlog("init_printing=%s", init_printing)
            init_printing(self.send_printers)
        except Exception as e:
            printlog.error("Error initializing printing support:")
            printlog.estr(e)
            self.printing = False
        else:
            self.send_printers()
        printlog("init_printing() enabled=%s", self.printing)

    def cleanup_printing(self) -> None:
        printlog("cleanup_printing() printing=%s", self.printing)
        if not self.printing:
            return
        self.cancel_send_printers_timer()
        try:
            from xpra.platform.printing import cleanup_printing  # pylint: disable=import-outside-toplevel
            printlog("cleanup_printing=%s", cleanup_printing)
            cleanup_printing()
        except ImportError:
            printlog("cleanup_printing()", exc_info=True)
        except Exception as e:
            printlog("cleanup_printing()", exc_info=True)
            printlog.warn("Warning: failed to cleanup printing subsystem:")
            printlog.warn(" %s", e)

    def send_printers(self, *args) -> None:
        printlog("send_printers%s timer=%s", args, self.send_printers_timer)
        # dbus can fire dozens of times for a single printer change,
        # so we wait a bit and fire via a timer to try to batch things together:
        if self.send_printers_timer:
            return
        self.send_printers_timer = GLib.timeout_add(500, self.do_send_printers)

    def cancel_send_printers_timer(self) -> None:
        spt = self.send_printers_timer
        printlog("cancel_send_printers_timer() send_printers_timer=%s", spt)
        if spt:
            self.send_printers_timer = 0
            GLib.source_remove(spt)

    def do_send_printers(self) -> None:
        self.send_printers_timer = 0
        start_thread(self.send_printers_thread, "send-printers", True)

    def send_printers_thread(self) -> None:
        from xpra.platform.printing import get_printers, get_mimetypes  # pylint: disable=import-outside-toplevel
        try:
            printers = get_printers()
        except Exception as e:
            printlog("%s", get_printers, exc_info=True)
            printlog.error("Error: cannot access the list of printers")
            printlog.estr(e)
            return
        printlog("send_printers_thread() found printers=%s", printers)
        try:
            # remove xpra-forwarded printers to avoid loops and multi-forwards,
            # also ignore stopped printers
            # and only keep the attributes that the server cares about (self.printer_attributes)
            exported_printers = {}

            def used_attrs(d):
                # filter attributes so that we only compare things that are actually used
                if not d:
                    return d
                return {pk: pv for pk, pv in d.items() if pk in self.printer_attributes}

            for k, v in printers.items():
                device_uri = v.get("device-uri", "")
                if device_uri:
                    # this is specific to the `cups` backend:
                    printlog("send_printers_thread() device-uri(%s)=%s", k, device_uri)
                    if device_uri.startswith("xpraforwarder"):
                        printlog("do_send_printers() skipping xpra forwarded printer=%s", k)
                        continue
                state = v.get("printer-state")
                # "3" if the destination is idle,
                # "4" if the destination is printing a job,
                # "5" if the destination is stopped.
                if state == 5 and SKIP_STOPPED_PRINTERS:
                    printlog("do_send_printers() skipping stopped printer=%s", k)
                    continue
                attrs = used_attrs(v)
                # add mimetypes:
                attrs["mimetypes"] = get_mimetypes()
                exported_printers[k] = attrs
            if self.exported_printers is None:
                # not been sent yet, ensure we can use the dict below:
                self.exported_printers = {}
            elif exported_printers == self.exported_printers:
                printlog("send_printers_thread() exported printers unchanged: %s", self.exported_printers)
                return
            # show summary of what has changed:
            added = tuple(k for k in exported_printers if k not in self.exported_printers)
            if added:
                printlog("send_printers_thread() new printers: %s", added)
            removed = tuple(k for k in self.exported_printers if k not in exported_printers)
            if removed:
                printlog("send_printers_thread() printers removed: %s", removed)
            modified = tuple(k for k, v in exported_printers.items() if
                             self.exported_printers.get(k) != v and k not in added)
            if modified:
                printlog("send_printers_thread() printers modified: %s", modified)
            printlog("send_printers_thread() printers=%s", exported_printers.keys())
            printlog("send_printers_thread() exported printers=%s", csv(str(x) for x in exported_printers))
            self.exported_printers = exported_printers
            self.send("printers", self.exported_printers)
        except Exception as e:
            printlog("do_send_printers()", exc_info=True)
            printlog.error("Error sending the list of printers")
            printlog.estr(e)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

MODIFIER_MAP = {
    "shift": 1 << 0,
    "lock": 1 << 1,
    "control": 1 << 2,
    "mod1": 1 << 3,
    "mod2": 1 << 4,
    "mod3": 1 << 5,
    "mod4": 1 << 6,
    "mod5": 1 << 7,
}

DEFAULT_MODIFIER_NAMES = ["shift", "control"]
DEFAULT_MODIFIER_NUISANCE_KEYNAMES = ["Num_Lock", "Caps_Lock", "Scroll_Lock"]
DEFAULT_MODIFIER_NUISANCE = ["lock"]
DEFAULT_ALL_MODIFIER_NAMES = DEFAULT_MODIFIER_NAMES + DEFAULT_MODIFIER_NUISANCE + ["mod1", "mod2", "mod3", "mod4",
                                                                                   "mod5"]

DEFAULT_MODIFIER_MEANINGS: dict[str, str] = {
    "Shift_L": "shift",
    "Shift_R": "shift",
    "Caps_Lock": "lock",
    "Control_L": "control",
    "Control_R": "control",
    "Alt_L": "mod1",
    "Alt_R": "mod1",
    "Meta_L": "mod1",
    "Meta_R": "mod1",
    "Num_Lock": "mod2",
    "Super_L": "mod3",
    "Super_R": "mod3",
    "Hyper_L": "mod4",
    "Hyper_R": "mod4",
    "ISO_Level3_Shift": "mod5",
    "Mode_switch": "mod5",
}


def mask_to_names(mask: int, modifier_map: dict[str, int]) -> list[str]:
    modifiers = []
    for modifier, modifier_mask in modifier_map.items():
        if modifier_mask & mask:
            modifiers.append(modifier)
            mask &= ~modifier_mask
    return modifiers



# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

import objc
from Cocoa import (
    NSOpenGLContext, NSOpenGLPixelFormat, NSOpenGLPFAAccelerated,
    NSOpenGLPFAAlphaSize, NSOpenGLPFAColorSize,
    NSOpenGLPFADepthSize, NSOpenGLPFADoubleBuffer, NSOpenGLPFAAccumSize,
    NSOpenGLPFAStencilSize, NSOpenGLPFAAuxBuffers, NSOpenGLCPSurfaceOpacity,
    NSOpenGLGetVersion,
    NSOpenGLPFAOpenGLProfile, NSOpenGLProfileVersion3_2Core,
)
from xpra.gtk.window import GDKWindow
from xpra.opengl.check import check_PyOpenGL_support
from xpra.platform.darwin.gdk3_bindings import get_nsview_ptr, enable_transparency, get_backing_scale_factor
from xpra.util.env import envbool
from xpra.os_util import gi_import
from xpra.log import Logger

log = Logger("opengl")

# require hardware accelerated OpenGL:
ACCELERATED = envbool("XPRA_OPENGL_ACCELERATED", False)


class AGLWindowContext:

    def __init__(self, gl_context: NSOpenGLContext, nsview: int, gdk_window):
        self.gl_context = gl_context
        self.nsview = nsview
        self.gdk_window = gdk_window
        log("%s", self)
        self.gl_context.setView_(nsview)

    def __enter__(self):
        assert self.gl_context
        self.gl_context.makeCurrentContext()
        return self

    def __exit__(self, *_args):
        NSOpenGLContext.clearCurrentContext()

    def swap_buffers(self) -> None:
        assert self.gl_context
        self.gl_context.flushBuffer()

    def update_geometry(self) -> None:
        """
        The window has been resized,
        the gl context must be updated.
        """
        glc = self.gl_context
        log("update_geometry() gl_context=%s", glc)
        if glc:
            glc.update()

    def __del__(self):
        self.destroy()

    def get_scale_factor(self) -> float:
        return get_backing_scale_factor(self.gdk_window)

    def destroy(self) -> None:
        self.gl_context = None
        self.nsview = 0

    def __repr__(self):
        return f"AGLWindowContext({self.gl_context}, {self.nsview})"


class AGLContext:

    def __init__(self, alpha=True):
        self.alpha = alpha
        self.gl_context: NSOpenGLContext | None = None
        self.nsview_ptr: int = 0
        self.window_context: AGLWindowContext | None = None
        attrs = []
        if ACCELERATED:
            attrs.append(NSOpenGLPFAAccelerated)
        attrs += [
            NSOpenGLPFADoubleBuffer,
            NSOpenGLPFADepthSize, 24,
            NSOpenGLPFAOpenGLProfile, NSOpenGLProfileVersion3_2Core,
        ]
        if alpha:
            attrs += [
                NSOpenGLPFAAlphaSize,
                8,
            ]
        attrs.append(0)
        log(f"AGLContext({alpha}) creating NSOpenGLPixelFormat from {attrs}")
        self.pixel_format = NSOpenGLPixelFormat.alloc().initWithAttributes_(attrs)
        assert self.pixel_format is not None, "failed to initialize NSOpenGLPixelFormat with {}".format(attrs)
        c = NSOpenGLContext.alloc().initWithFormat_shareContext_(self.pixel_format, None)
        assert c is not None, "failed to initialize NSOpenGLContext with {}".format(self.pixel_format)
        self.gl_context = c

    def check_support(self, force_enable: bool = False) -> dict[str, Any]:
        # map our names (based on GTK's) to apple's constants:
        attr_name = {
            "rgba": (bool, NSOpenGLPFAAlphaSize),
            "depth": (int, NSOpenGLPFAColorSize),
            # "red-size"          : ?
            # "green-size"        : ?
            # "blue-size"         : ?
            # "red-shift"         : ?
            # "green-shift"       : ?
            # "blue-shift"        : ?
            # "alpha-shift"       : ?
            # "accum-red-size"    : ?
            # "accum-green-size"  : ?
            # "accum-blue-size"   : ?
            "alpha-size": (int, NSOpenGLPFAAlphaSize),
            "accum-size": (int, NSOpenGLPFAAccumSize),
            "depth-size": (int, NSOpenGLPFADepthSize),
            "stencil-size": (int, NSOpenGLPFAStencilSize),
            "aux-buffers": (int, NSOpenGLPFAAuxBuffers),
            # "visible-mask"      : ?
            "double-buffered": (int, NSOpenGLPFADoubleBuffer)
        }
        major, minor = NSOpenGLGetVersion(None, None)
        log(f"NSOpenGLGetVersion()={major},{minor}")
        nscreens = self.pixel_format.numberOfVirtualScreens()
        i = {
            # "pixel-format"      : self.pixel_format,
            "virtual-screens": nscreens,
        }
        for name, vdef in attr_name.items():
            conv, const_val = vdef  # ie (bool, NSOpenGLPFAAlphaSize)
            v = self._get_apfa(const_val)  # ie: NSOpenGLPFAAlphaSize=8
            i[name] = conv(v)  # ie: bool(8)
        # do it again but for each screen:
        if nscreens > 1:
            for screen in range(nscreens):
                si = i.setdefault("screen-%i" % screen, {})
                for name, vdef in attr_name.items():
                    conv, const_val = vdef  # ie (bool, NSOpenGLPFAAlphaSize)
                    v = self._get_pfa(const_val, screen)  # ie: NSOpenGLPFAAlphaSize=8
                    si[name] = conv(v)  # ie: bool(8)
        Gdk = gi_import("Gdk")
        tmp = GDKWindow(window_type=Gdk.WindowType.TEMP, title="tmp-opengl-check")
        with self.get_paint_context(tmp):
            i.update(check_PyOpenGL_support(force_enable))
        tmp.hide()
        return i

    def _get_pfa(self, attr, screen):
        return self.pixel_format.getValues_forAttribute_forVirtualScreen_(None, attr, screen)

    def _get_apfa(self, attr, fn=min):
        return fn(self._get_pfa(attr, screen) for screen in range(self.pixel_format.numberOfVirtualScreens()))

    def get_bit_depth(self) -> int:
        return int(self._get_apfa(NSOpenGLPFAColorSize))

    def is_double_buffered(self) -> bool:
        return bool(self._get_apfa(NSOpenGLPFADoubleBuffer))

    def get_paint_context(self, gdk_window) -> AGLWindowContext:
        if not self.gl_context:
            raise RuntimeError("no OpenGL context")
        nsview_ptr = get_nsview_ptr(gdk_window)
        if self.window_context and self.nsview_ptr != nsview_ptr:
            log("get_paint_context(%s) nsview_ptr has changed, was %#x, now %#x - destroying window context",
                gdk_window, nsview_ptr, self.nsview_ptr)
            self.window_context.destroy()
            self.window_context = None
        if not self.window_context:
            self.nsview_ptr = nsview_ptr
            nsview = objc.objc_object(c_void_p=nsview_ptr)
            log("get_paint_context(%s) nsview(%#x)=%s", gdk_window, nsview_ptr, nsview)
            if self.alpha and enable_transparency:
                self.gl_context.setValues_forParameter_([0], NSOpenGLCPSurfaceOpacity)
                enable_transparency(gdk_window)
            self.window_context = AGLWindowContext(self.gl_context, nsview, gdk_window)
        return self.window_context

    def __del__(self):
        self.destroy()

    def destroy(self) -> None:
        c = self.window_context
        log("AGLContext.destroy() window_context=%s", c)
        if c:
            self.window_context = None
            c.destroy()
        glc = self.gl_context
        if glc:
            self.gl_context = None
            glc.clearDrawable()
        self.pixel_format = None

    def __repr__(self):
        return "AGLContext(%#x)" % (self.nsview_ptr or 0)


GLContext = AGLContext



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import struct
from typing import Any, cast
from collections.abc import Callable, Sequence

from xpra.util.str_fn import Ellipsizer, bytestostr, hexstr
from xpra.log import Logger
log = Logger("x11")


class Unmanageable(Exception):
    pass


REPR_FUNCTIONS: dict[type, Callable] = {}


# Just to make it easier to pass around and have a helpful debug logging.
# Really, just a python objects where we can stick random bags of attributes.
class X11Event:
    def __init__(self, name: str):
        self.name: str = name

    def __repr__(self):
        d = {}
        for k, v in self.__dict__.items():
            if k in ("name", "display", "type"):
                continue
            if k in ("serial", "window", "delivered_to", "above", "below", "damage", "time") and isinstance(v, int):
                d[k] = hex(v)
            elif k == "send_event" and v is False:
                continue
            else:
                fn = REPR_FUNCTIONS.get(type(v), str)
                d[k] = fn(v)
        return f"<X11:{self.name} {d!r}>"


# we duplicate some of the code found in xpra.x11.gtk.prop ...
# which is still better than having dependencies on that GTK here
def get_X11_window_property(xid: int, name: str, req_type: str):
    try:
        from xpra.x11.bindings.window import X11WindowBindings, PropertyError
        try:
            prop = X11WindowBindings().XGetWindowProperty(xid, name, req_type)
            log("get_X11_window_property(%#x, %s, %s)=%s, len=%s", xid, name, req_type, type(prop), len(prop or ()))
            return prop
        except PropertyError as e:
            log("get_X11_window_property(%#x, %s, %s): %s", xid, name, req_type, e)
    except Exception as e:
        log.warn(f"Warning: failed to get X11 window property {name!r} on window {xid:x}: {e}")
        log("get_X11_window_property%s", (xid, name, req_type), exc_info=True)
    return None


def get_X11_root_property(name: str, req_type: str):
    try:
        from xpra.x11.bindings.window import X11WindowBindings
        return get_X11_window_property(X11WindowBindings().get_root_xid(), name, req_type)
    except Exception as e:
        log("get_X11_root_property(%s, %s)", name, req_type, exc_info=True)
        log.warn(f"Warning: failed to get X11 root property {name!r}")
        log.warn(" %s", e)
    return None


def get_wm_name() -> str:
    try:
        wm_check = get_X11_root_property("_NET_SUPPORTING_WM_CHECK", "WINDOW")
        if wm_check:
            xid = struct.unpack(b"@L", wm_check)[0]
            log("_NET_SUPPORTING_WM_CHECK window=%#x", xid)
            wm_name = get_X11_window_property(xid, "_NET_WM_NAME", "UTF8_STRING")
            log("_NET_WM_NAME=%s", wm_name)
            if wm_name:
                return wm_name.decode("utf8")
    except Exception as e:
        log("get_wm_name()", exc_info=True)
        log.error("Error accessing window manager information:")
        log.estr(e)
    return ""


def get_icc_data() -> dict[str, Any]:
    icc: dict[str, Any] = {}
    try:
        data = get_X11_root_property("_ICC_PROFILE", "CARDINAL")
        if data:
            log("_ICC_PROFILE=%s (%s)", type(data), len(data))
            version = get_X11_root_property("_ICC_PROFILE_IN_X_VERSION", "CARDINAL")
            log("get_icc_info() found _ICC_PROFILE_IN_X_VERSION=%s, _ICC_PROFILE=%s",
                hexstr(version or ""), Ellipsizer(hexstr(data)))
            icc |= {
                "source": "_ICC_PROFILE",
                "data": data,
            }
            if version:
                try:
                    version = ord(version)
                except TypeError:
                    pass
                icc["version"] = version
    except Exception as e:
        log.error("Error: cannot access `_ICC_PROFILE` X11 window property")
        log.estr(e)
        log("get_icc_info()", exc_info=True)
    log("get_x11_icc_data()=%s", Ellipsizer(icc))
    return icc


def get_current_desktop() -> int:
    v = -1
    d = None
    try:
        d = get_X11_root_property("_NET_CURRENT_DESKTOP", "CARDINAL")
        if d:
            v = struct.unpack(b"@L", d)[0]
    except Exception as e:
        log("get_current_desktop()", exc_info=True)
        log.error("Error: accessing `_NET_CURRENT_DESKTOP`:")
        log.estr(e)
    log("get_current_desktop() %s=%s", hexstr(d or ""), v)
    return v


def get_number_of_desktops() -> int:
    v = 0
    d = None
    try:
        d = get_X11_root_property("_NET_NUMBER_OF_DESKTOPS", "CARDINAL")
        if d:
            v = struct.unpack(b"@L", d)[0]
    except Exception as e:
        log("get_number_of_desktops()", exc_info=True)
        log.error("Error: accessing `_NET_NUMBER_OF_DESKTOPS`:")
        log.estr(e)
    v = max(1, v)
    log("get_number_of_desktops() %s=%s", hexstr(d or ""), v)
    return v


def get_workarea() -> tuple[int, int, int, int] | None:
    try:
        d = get_current_desktop()
        if d < 0:
            return None
        workarea = get_X11_root_property("_NET_WORKAREA", "CARDINAL")
        if not workarea:
            return None
        log("get_workarea() _NET_WORKAREA=%s (%s), len=%s",
            Ellipsizer(workarea), type(workarea), len(workarea))
        # workarea comes as a list of 4 CARDINAL dimensions (x,y,w,h), one for each desktop
        sizeof_long = struct.calcsize(b"@L")
        if len(workarea) < (d+1)*4*sizeof_long:
            log.warn(f"Warning: invalid `_NET_WORKAREA` value length: {workarea!r}")
        else:
            cur_workarea = workarea[d*4*sizeof_long:(d+1)*4*sizeof_long]
            v = cast(tuple[int, int, int, int], struct.unpack(b"@LLLL", cur_workarea))
            log("get_workarea() %s=%s", hexstr(cur_workarea), v)
            return v
    except Exception as e:
        log("get_workarea()", exc_info=True)
        log.error("Error: querying the x11 workarea:")
        log.estr(e)
    return None


def get_desktop_names() -> Sequence[str]:
    ret: Sequence[str] = ("Main", )
    d = None
    try:
        d = get_X11_root_property("_NET_DESKTOP_NAMES", "UTF8_STRING")
        if d:
            d = d.split(b"\0")
            if len(d) > 1 and d[-1] == b"":
                d = d[:-1]
            ret = tuple(x.decode("utf8") for x in d)
    except Exception as e:
        log.error("Error querying `_NET_DESKTOP_NAMES`:")
        log.estr(e)
    log("get_desktop_names() %s=%s", hexstr(d or ""), ret)
    return ret


def get_vrefresh() -> int:
    v = -1
    try:
        from xpra.x11.bindings.randr import RandRBindings
        randr = RandRBindings()
        if randr.has_randr():
            v = randr.get_vrefresh()
    except Exception as e:
        log("get_vrefresh()", exc_info=True)
        log.error("Error querying the display vertical refresh rate:")
        log.estr(e)
    log("get_vrefresh()=%s", v)
    return v


def send_client_message(window, message_type: str, *values) -> None:
    try:
        from xpra.x11.bindings.window import constants, X11WindowBindings  # @UnresolvedImport
        X11Window = X11WindowBindings()
        root_xid = X11Window.get_root_xid()
        if window:
            xid = window.get_xid()
        else:
            xid = root_xid
        SubstructureNotifyMask = constants["SubstructureNotifyMask"]
        SubstructureRedirectMask = constants["SubstructureRedirectMask"]
        event_mask = SubstructureNotifyMask | SubstructureRedirectMask
        X11Window.sendClientMessage(root_xid, xid, False, event_mask, message_type, *values)
    except Exception as e:
        log.warn(f"Warning: failed to send client message {message_type!r} with values={values}: {e}")


device_bell = None


def system_bell(xid: int, device, percent: int, _pitch, _duration: int, bell_class, bell_id, bell_name: str) -> bool:
    global device_bell
    if device_bell is False:
        # failed already
        return False
    if device_bell is None:
        # try to load it:
        try:
            from xpra.x11.bindings.keyboard import X11KeyboardBindings
            device_bell = X11KeyboardBindings().device_bell
        except ImportError:
            log("x11_bell()", exc_info=True)
            log.warn("Warning: cannot use X11 bell device without the X11 bindings")
            return False
    device_bell(xid, device, bell_class, bell_id, percent, bell_name)
    return True


def get_xsettings():
    from xpra.x11.bindings.window import X11WindowBindings  # @UnresolvedImport
    X11Window = X11WindowBindings()
    selection = "_XSETTINGS_S0"
    owner = X11Window.XGetSelectionOwner(selection)
    if not owner:
        return None
    XSETTINGS = "_XSETTINGS_SETTINGS"
    data = X11Window.XGetWindowProperty(owner, XSETTINGS, XSETTINGS)
    if not data:
        return None
    from xpra.x11.xsettings_prop import bytes_to_xsettings
    return bytes_to_xsettings(data)


def xsettings_to_dict(v) -> dict[str, tuple[int, Any]]:
    d: dict[str, tuple[int, Any]] = {}
    if v:
        _, values = v
        for setting_type, prop_name, value, _ in values:
            d[bytestostr(prop_name)] = (setting_type, value)
    return d


def get_randr_dpi() -> tuple[int, int]:
    from xpra.x11.bindings.randr import RandRBindings  # @UnresolvedImport
    randr_bindings = RandRBindings()
    if randr_bindings and randr_bindings.has_randr():
        wmm, hmm = randr_bindings.get_screen_size_mm()
        if wmm > 0 and hmm > 0:
            w, h = randr_bindings.get_screen_size()
            dpix = round(w * 25.4 / wmm)
            dpiy = round(h * 25.4 / hmm)
            log("xdpi=%s, ydpi=%s - size-mm=%ix%i, size=%ix%i", dpix, dpiy, wmm, hmm, w, h)
            return dpix, dpiy
    return -1, -1


def get_xresources() -> dict[str, str] | None:
    try:
        value = get_X11_root_property("RESOURCE_MANAGER", "STRING")
        log(f"RESOURCE_MANAGER={value}")
        if value is None:
            return None
        # parse the resources into a dict:
        values: dict[str, str] = {}
        options = bytestostr(value).split("\n")
        for option in options:
            if not option:
                continue
            parts = option.split(":\t", 1)
            if len(parts) != 2:
                log(f"skipped invalid option: {option!r}")
                continue
            values[parts[0]] = parts[1]
        return values
    except Exception as e:
        log(f"_get_xresources error: {e!r}")
    return None


def get_cursor_size() -> int:
    d = get_xresources() or {}
    try:
        return int(d.get("Xcursor.size", 0))
    except ValueError:
        return -1



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys

from xpra.util.parsing import parse_str_dict
from xpra.auth.common import SessionData, parse_uid, parse_gid
from xpra.auth.sys_auth_base import log
from xpra.auth.sqlauthbase import SQLAuthenticator, DatabaseUtilBase, run_dbutil


class Authenticator(SQLAuthenticator):

    def __init__(self, filename="sqlite.sdb", **kwargs):
        super().__init__(**kwargs)
        if filename and not os.path.isabs(filename):
            exec_cwd = kwargs.get("exec_cwd", os.getcwd())
            filename = os.path.join(exec_cwd, filename)
        self.filename = filename
        self.password_query = kwargs.pop("password_query", "SELECT password FROM users WHERE username=(?)")
        self.sessions_query = kwargs.pop("sessions_query",
                                         "SELECT uid, gid, displays, env_options, session_options "
                                         "FROM users WHERE username=(?) AND password=(?)")
        self.authenticate_check = self.authenticate_hmac

    def __repr__(self):
        return "sqlite"

    def db_cursor(self, *sqlargs):
        if not os.path.exists(self.filename):
            log.error("Error: sqlauth cannot find the database file '%s'", self.filename)
            return None
        import sqlite3  # pylint: disable=import-outside-toplevel
        db = sqlite3.connect(self.filename)
        db.row_factory = sqlite3.Row
        cursor = db.cursor()
        cursor.execute(*sqlargs)
        log("db_cursor(%s)=%s", sqlargs, cursor)
        return cursor

    def parse_session_data(self, data) -> SessionData | None:
        try:
            uid = parse_uid(data["uid"])
            gid = parse_gid(data["gid"])
            displays: list[str] = []
            env_options: dict[str, str] = {}
            session_options: dict[str, str] = {}
            if data["displays"]:
                displays = [x.strip() for x in str(data["displays"]).split(",") if x.strip()]
            if data["env_options"]:
                env_options = parse_str_dict(str(data["env_options"]), ";")
            if data["session_options"]:
                session_options = parse_str_dict(str(data["session_options"]), ";")
        except Exception as e:
            log("get_sessions() error on row %s", data, exc_info=True)
            log.error("Error: sqlauth database row parsing problem:")
            log.estr(e)
            return None
        return uid, gid, displays, env_options, session_options


class SqliteDatabaseUtil(DatabaseUtilBase):

    def __init__(self, uri):
        super().__init__(uri)
        import sqlite3  # pylint: disable=import-outside-toplevel
        assert sqlite3.paramstyle == "qmark"
        self.param = "?"

    def exec_database_sql_script(self, cursor_cb, *sqlargs):
        import sqlite3  # pylint: disable=import-outside-toplevel
        db = sqlite3.connect(self.uri)
        cursor = db.cursor()
        log("%s.execute%s", cursor, sqlargs)
        cursor.execute(*sqlargs)
        if cursor_cb:
            cursor_cb(cursor)
        db.commit()
        return cursor

    def get_authenticator_class(self) -> type:
        return Authenticator


def main(argv) -> int:
    return run_dbutil(SqliteDatabaseUtil, "filename", argv)


if __name__ == "__main__":
    sys.exit(main(sys.argv))



# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any
from collections.abc import Sequence

from xpra.gstreamer.common import (
    GST_FLOW_OK, STREAM_TYPE, GST_FORMAT_BYTES,
    make_buffer, has_plugins,
    get_caps_str, get_element_str,
)
from xpra.codecs.gstreamer.common import (
    VideoPipeline,
    get_version, get_type, get_info,
    get_default_decoder_options,
)
from xpra.codecs.constants import VideoSpec
from xpra.codecs.image import ImageWrapper
from xpra.common import roundup
from xpra.os_util import WIN32, gi_import
from xpra.util.objects import typedict
from xpra.log import Logger

log = Logger("decoder", "gstreamer")

GObject = gi_import("GObject")

log(f"decoder: {get_type()} {get_version()}")

FORMATS = os.environ.get("XPRA_GSTREAMER_DECODER_FORMATS", "h264,hevc,vp8,vp9,av1").split(",")


def get_default_mappings() -> dict[str, Sequence[str]]:
    # should always be available:
    m: dict[str, Sequence[str]] = {
        "vp8": ("vp8dec",),
        "vp9": ("vp9dec",),
    }
    if WIN32:
        m["h264"] = ("d3d11h264dec",)
    else:
        m["av1"] = ("av1dec",)
        # enable nv decoder unless we don't find nvidia hardware:
        h264 = ["nvh264dec"]
        try:
            from xpra.codecs.nvidia.util import has_nvidia_hardware
            if not has_nvidia_hardware():
                h264 = []
        except ImportError:
            pass
        h264.append("avdec_h264")
        m["h264"] = tuple(h264)
        m["hevc"] = ("vaapih265dec",)
    return m


def get_codecs_options() -> dict[str, Sequence[str]]:
    dm = os.environ.get("XPRA_GSTREAMER_DECODER_MAPPINGS")
    if not dm:
        return get_default_mappings()
    codec_options = {}
    for mapping in dm.split(";"):  # ie: mapping="vp8:vp8dec"
        try:
            enc, elements_str = mapping.split(":", 1)
        except IndexError:
            log.warn(f"Warning: invalid decoder mapping {mapping}")
        else:
            # ie: codec_options["h264"] = ["avdec_h264", "nvh264dec"]
            codec_options[enc] = tuple(elements_str.split(","))
    return codec_options


def find_codecs(options) -> dict[str, str]:
    codecs: dict[str, str] = {}
    for encoding, elements in options.items():
        if encoding in FORMATS and elements:
            found = [x for x in elements if has_plugins(x)]
            if found:
                codecs[encoding] = found[0]
    log(f"find_codecs({options})={codecs}")
    return codecs


CODECS = find_codecs(get_codecs_options())


def get_encodings() -> Sequence[str]:
    return tuple(CODECS.keys())


def get_min_size(_encoding: str):
    return 48, 16


def get_specs() -> Sequence[VideoSpec]:
    specs: Sequence[VideoSpec] = []
    for encoding, decoder in CODECS.items():
        in_cs = "YUV420P"
        out_cs = "NV12" if decoder.startswith("nv") else "YUV420P"
        has_lossless_mode = encoding == "vp9" and in_cs == "YUV444P"
        specs.append(VideoSpec(
            encoding=encoding,
            input_colorspace=in_cs, output_colorspaces=(out_cs, ),
            has_lossless_mode=has_lossless_mode,
            codec_class=Decoder, codec_type=get_type(),
            quality=50, speed=50,
            size_efficiency=60,
            setup_cost=50,
            max_w=8192,
            max_h=4096,
        ))
    return specs


class Decoder(VideoPipeline):
    __gsignals__: dict[str, tuple] = VideoPipeline.__generic_signals__.copy()
    decoder_element = "unset"
    """
    Dispatch video decoding to a gstreamer pipeline
    """

    def __repr__(self):
        if self.colorspace is None:
            return f"gstreamer-{self.decoder_element}(uninitialized)"
        return f"gstreamer-{self.decoder_element}({self.colorspace} - {self.width}x{self.height})"

    def create_pipeline(self, options: typedict) -> None:
        if self.encoding not in get_encodings():
            raise ValueError(f"invalid encoding {self.encoding!r}")
        self.dst_formats = options.strtupleget("dst-formats")
        self.decoder_element = CODECS.get(self.encoding)
        if not self.decoder_element:
            raise RuntimeError(f"invalid encoding {self.encoding}")
        stream_attrs: dict[str, Any] = {
            "width": self.width,
            "height": self.height,
        }
        eopts = get_default_decoder_options().get(self.decoder_element, {})
        if not eopts:
            eopts = {
                "profile": "main",
                "stream-format": "byte-stream",
                "alignment": "au",
            }
        for k, v in eopts.items():
            stream_attrs[k] = options.strget(k, v)
        stream_caps = get_caps_str(f"video/x-{self.encoding}", stream_attrs)
        if self.decoder_element.startswith("nv"):
            gst_format = "NV12"
            self.output_format = "NV12"
        else:
            gst_format = "I420"
            self.output_format = "YUV420P"
        output_caps = get_caps_str("video/x-raw", {
            "width": self.width,
            "height": self.height,
            "format": gst_format,
        })
        elements = [
            get_element_str("appsrc", {
                "name": "src",
                "emit-signals": 1,
                "block": 0,
                "is-live": 1,
                "do-timestamp": 1,
                "stream-type": STREAM_TYPE,
                "format": GST_FORMAT_BYTES,
                "caps": stream_caps,
            }),
            f"{self.decoder_element} name=decoder",
            get_element_str("appsink", {
                "name": "sink",
                "emit-signals": 1,
                "max-buffers": 10,
                "drop": False,
                "sync": False,
                "async": True,
                "qos": False,
                "caps": output_caps,
            })
        ]
        if not self.setup_pipeline_and_bus(elements):
            raise RuntimeError("failed to setup gstreamer pipeline")

    def get_colorspace(self) -> str:
        return self.colorspace

    def on_new_sample(self, _bus) -> int:
        sample = self.sink.emit("pull-sample")
        buf = sample.get_buffer()
        size = buf.get_size()
        log("on_new_sample size=%s, output_format=%s", size, self.output_format)
        if size:
            mem = memoryview(buf.extract_dup(0, size))
            # I420 gstreamer definition:
            Ystride = roundup(self.width, 4)
            Ysize = Ystride * roundup(self.height, 2)
            Y = mem[:Ysize]
            planes: Sequence[memoryview]
            strides: Sequence[int]
            if self.output_format == "YUV420P":
                UVstride = roundup(roundup(self.width, 2) // 2, 4)
                UVsize = UVstride * roundup(self.height, 2) // 2
                total = Ysize + 2 * UVsize
                if size < total:
                    raise RuntimeError(f"I420 sample buffer is too small: expected {total} but got {size}")
                U = mem[Ysize:Ysize + UVsize]
                V = mem[Ysize + UVsize:total]
                planes = (Y, U, V)
                strides = (Ystride, UVstride, UVstride)
            else:
                UVstride = roundup(self.width, 4)
                UVsize = UVstride * roundup(self.height, 2) // 2
                UV = mem[Ysize:Ysize + UVsize]
                planes = (Y, UV)
                strides = (Ystride, UVstride)
            image = ImageWrapper(0, 0, self.width, self.height, planes,
                                 self.output_format, 24, strides, 3, ImageWrapper.PLANAR_3)
            self.frame_queue.put(image)
        return GST_FLOW_OK

    def decompress_image(self, data: bytes, options: typedict) -> ImageWrapper | None:
        log(f"decompress_image(.., {options}) state={self.state} data size={len(data)}")
        if self.state in ("stopped", "error"):
            log(f"pipeline is in {self.state} state, dropping buffer")
            return None
        buf = make_buffer(data)
        # duration = normv(0)
        # if duration>0:
        #    buf.duration = duration
        # buf.size = size
        # buf.timestamp = timestamp
        # buf.offset = offset
        # buf.offset_end = offset_end
        return self.process_buffer(buf, options)


GObject.type_register(Decoder)


def selftest(full=False) -> None:
    log("gstreamer decoder selftest: %s", get_info())
    from xpra.codecs.checks import testdecoder
    from xpra.codecs.gstreamer import decoder
    remaining = testdecoder(decoder, full)
    decoder.CODECS = {k: v for k, v in decoder.CODECS.items() if k in remaining}



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from collections.abc import Sequence

from xpra.log import Logger
from xpra.util.env import envbool
from xpra.util.str_fn import csv
from xpra.exit_codes import ExitCode
from xpra.scripts.config import InitExit


GSTREAMER_CAPTURE_ELEMENTS: Sequence[str] = ("ximagesrc", "pipewiresrc")

XSHM: bool = envbool("XPRA_SHADOW_XSHM", True)
NVFBC: bool = envbool("XPRA_SHADOW_NVFBC", True)
GSTREAMER: bool = envbool("XPRA_SHADOW_GSTREAMER", True)
PIPEWIRE: bool = envbool("XPRA_SHADOW_PIPEWIRE", True)


def warn(*messages) -> None:
    log = Logger("server", "shadow")
    log("warning loading backend", exc_info=True)
    for m in messages:
        log.warn(m)


def debug_error(message: str) -> None:
    log = Logger("server", "shadow")
    log(message, exc_info=True)


def load_screencast(display: str = "") -> type | None:
    try:
        from xpra.platform.posix import screencast
        return screencast.ScreenCast
    except ImportError as e:
        debug_error(f"load_screencast({display})")
        warn("Warning: unable to load the screencast backend", f" {e}")
    return None


def load_remotedesktop(display: str = "") -> type | None:
    try:
        from xpra.platform.posix import remotedesktop
        return remotedesktop.RemoteDesktop
    except ImportError as e:
        debug_error(f"load_remotedesktop({display})")
        warn("Warning: unable to load the remotedesktop backend", f" {e}")
    return None


def load_pipewire(display: str = "") -> type | None:
    return load_remotedesktop() or load_screencast()


def load_wayland(display: str = "") -> type | None:
    c = load_remotedesktop() or load_screencast()
    if c:
        os.environ["GDK_BACKEND"] = "wayland"
        if display:
            os.environ["WAYLAND_DISPLAY"] = display
        if os.environ.get("XPRA_NOX11") is None:
            os.environ["XPRA_NOX11"] = "1"
    return c


def load_x11(display: str = "") -> type | None:
    gdkb = os.environ.get("GDK_BACKEND")
    try:
        os.environ["GDK_BACKEND"] = "x11"
        from xpra.x11.server import shadow
        return shadow.ShadowX11Server
    except ImportError as e:
        debug_error(f"load_x11({display})")
        if gdkb:
            os.environ["GDK_BACKEND"] = gdkb
        warn("Warning: unable to load x11 shadow server", f" {e}")
    return None


# the ShadowX11Server supports multiple sub-backends:
load_nvfbc = load_x11
load_xshm = load_x11
load_gtk = load_x11


def load_auto(display: str = "") -> type | None:
    c: type | None = None
    if display.startswith("wayland-") or os.path.isabs(display):
        c = load_wayland(display)
    elif display.startswith(":"):
        c = load_x11(display)
    return c or load_remotedesktop(display) or load_screencast(display) or load_x11(display)


def ShadowServer(display: str, attrs: dict[str, str]):
    log = Logger("server", "shadow")
    setting = (attrs.get("backend", os.environ.get("XPRA_SHADOW_BACKEND", "auto"))).lower()
    log(f"ShadowServer({display}, {attrs}) {setting=}")
    if setting not in SHADOW_OPTIONS:
        raise InitExit(ExitCode.UNSUPPORTED, f"invalid shadow backend {setting!r}, use: {csv(SHADOW_OPTIONS.keys())}")
    load_fn = globals().get(f"load_{setting}")
    if not load_fn:
        raise RuntimeError(f"missing shadow loader for {setting!r}")
    shadow_server = load_fn(display)
    if not shadow_server:
        raise RuntimeError(f"shadow backend {setting} is not available")
    log(f"ShadowServer({display}, {attrs}) {setting}={shadow_server}")
    return shadow_server(attrs)


def check_nvfbc() -> bool:
    return NVFBC


def check_gstreamer() -> bool:
    if not GSTREAMER:
        return False
    from xpra.gstreamer.common import has_plugins, import_gst
    import_gst()
    return has_plugins("ximagesrc")


def check_pipewire() -> bool:
    if not PIPEWIRE:
        return False
    from xpra.gstreamer.common import has_plugins, import_gst
    import_gst()
    return has_plugins("pipewiresrc")


def check_x11() -> bool:
    from xpra.x11.bindings.ximage import XImageBindings  # pylint: disable=import-outside-toplevel
    assert XImageBindings
    return True


def check_xshm() -> bool:
    from xpra.x11.bindings.ximage import XImageBindings  # pylint: disable=import-outside-toplevel
    assert XImageBindings
    return XImageBindings().has_XShm()


def nocheck() -> bool:
    return True


SHADOW_OPTIONS = {
    "auto": nocheck,
    "nvfbc": check_nvfbc,
    "gstreamer": check_gstreamer,
    "pipewire": check_pipewire,
    "x11": check_x11,
    "xshm": check_xshm,
    "gtk": check_x11,
}



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from xpra.log import Logger

log = Logger("command")


class ControlHandler:

    def __init__(self):
        self.control_commands: dict[str, Any] = {}

    def add_default_control_commands(self, enabled=True):
        # for things that can take longer:
        try:
            from xpra.net.control.common import HelloCommand, HelpCommand, DisabledCommand
            from xpra.net.control.debug import DebugControl
        except ImportError:
            return
        self.control_commands = {
            "hello": HelloCommand(),
        }
        if enabled:
            self.add_control_command("debug", DebugControl())
            self.add_control_command("help", HelpCommand(self.control_commands))
        else:
            self.add_control_command("*", DisabledCommand())

    def add_control_command(self, name: str, control) -> None:
        self.control_commands[name] = control

    def handle_command_request(self, proto, *args) -> None:
        """ client sent a command request as part of the hello packet """
        assert args, "no arguments supplied"
        from xpra.net.control.common import process_control_command
        code, response = process_control_command(proto, self.control_commands, *args)
        hello = {"command_response": (code, response)}
        proto.send_now(("hello", hello))



#!/usr/bin/env python3

# BATch files are useless,
# all we want is sed,
# but is is MSWindows so it's easier to re-invent the wheel, sigh

import sys

def main():
    if len(sys.argv)<3:
        print("Invalid number of arguments")
        print("Usage: %s source-msi.xml destination-msi.xml TOKEN1=VALUE1 [TOKEN2=VALUE2 [..]]", sys.argv[0])
        sys.exit(1)
    src = sys.argv[1]
    dst = sys.argv[2]
    print("replacing tokens from file '%s' and saving to '%s'" % (src, dst))
    data = open(src, 'rb').read().decode()
    print("data=%i bytes %s" % (len(data), type(data)))
    for kv in sys.argv[3:]:
        try:
            token, value = kv.split("=")
        except ValueError:
            print("skipping invalid token string '%s'" % kv)
            continue
        print("replacing %s with '%s' %s" % (token, value, type(token)))
        data = data.replace("$%s" % token, value)
        data = data.replace("${%s}" % token, value)
    open(dst, 'wb').write(data.encode())

if __name__ == '__main__':
    main()
